"""Configuration handling in json/yaml files."""

import os
from typing import Any, Optional

from pnp import validator
from pnp.config._base import Configuration, ConfigLoader
from pnp.config._yaml import YamlConfigLoader
from pnp.models import PullModel

# Available config loader
_LOADER = [YamlConfigLoader]


# Available configuration file types and their associated loader
_EXT_LOADER_MAP = {
    ext: clazz for clazz in _LOADER for ext in clazz.supported_extensions()
}


# Global variable: Which loader was used to load the config
_LOADER_USED = None  # type: Optional[ConfigLoader]


def load_config(config_path: str) -> Configuration:
    """Load the specified config by using a compatible `ConfigLoader`."""
    global _LOADER_USED  # pylint: disable=global-statement

    validator.is_file(config_path=config_path)

    _, ext = os.path.splitext(config_path)
    loader_clazz = _EXT_LOADER_MAP.get(ext.lower().lstrip('.'))
    if not loader_clazz:
        raise RuntimeError(
            "No configuration loader is able to load a '*{}' configuration file".format(ext)
        )
    _LOADER_USED = loader_clazz()
    return _LOADER_USED.load_config(config_path)


def load_pull_from_snippet(snippet: Any, name: str, **extra: Any) -> PullModel:
    """Loads a pull from a snippet using the same `ConfigLoader` that was initially used
    to load the main configuration."""
    global _LOADER_USED  # pylint: disable=global-statement
    loader = _LOADER_USED or YamlConfigLoader()  # Default loader for testing
    return loader.load_pull_from_snippet(snippet, name, **extra)



"""Home assistant related utility classes."""

import json
import urllib.parse as urlparse
from typing import Any, Optional

from typeguard import typechecked

from pnp.utils import ReprMixin


class HassApi(ReprMixin):
    """Utility class to communicate with home assistant via the rest-api."""

    METHOD_GET = 'get'
    METHOD_POST = 'post'
    ALLOWED_METHODS = [METHOD_GET, METHOD_POST]

    __REPR_FIELDS__ = ["base_url", "timeout"]

    @typechecked
    def __init__(self, base_url: str, token: str, timeout: Optional[int] = None):
        self.base_url = base_url
        self.token = token
        self.timeout = timeout and int(timeout)

    @typechecked
    def call(self, endpoint: str, method: str = METHOD_GET, data: Any = None) -> Any:
        """Calls the specified endpoint (without prefix api) using the given method.
        You can optionally pass data to the request which will be json encoded."""
        from requests import get, post

        method = str(method).lower()
        if method not in self.ALLOWED_METHODS:
            raise ValueError(
                "Argument method is expected to be one of {allowed}, but is '{method}'".format(
                    allowed=str(self.ALLOWED_METHODS),
                    method=method)
            )

        url = urlparse.urljoin(urlparse.urljoin(self.base_url, 'api/'), endpoint)
        headers = {
            'Authorization': 'Bearer {self.token}'.format(**locals()),
            'content-type': 'application/json',
        }

        if data is not None:
            data = json.dumps(data)

        if method == self.METHOD_GET:
            response = get(url, headers=headers, timeout=self.timeout, data=data)
        else:
            response = post(url, headers=headers, timeout=self.timeout, data=data)

        if response.status_code != 200:
            raise RuntimeError("Failed to call endpoint {url}"
                               "\nHttp Code: {response.status_code}"
                               "\nMessage: {response.text}".format(**locals()))

        return response.json()



"""Contains metrics related endpoints."""

from fastapi import FastAPI
from starlette_exporter import PrometheusMiddleware, handle_metrics

from .base import Endpoint


class PrometheusExporter(Endpoint):
    """Prometheus metrics exporter endpoint."""

    def __init__(self, app_name: str):
        self.app_name = app_name

    def attach(self, fastapi: FastAPI) -> None:
        fastapi.add_middleware(
            PrometheusMiddleware,
            app_name=self.app_name,
            group_paths=True
        )
        fastapi.add_route("/metrics", handle_metrics)



"""Publishing engine classes to the outside world."""

from pnp.engines._async import AsyncEngine
from pnp.engines._base import (
    Engine,
    RetryDirective,
    RetryHandler,
    NoRetryHandler,
    SimpleRetryHandler,
    LimitedRetryHandler,
    AdvancedRetryHandler,
    PushExecutor,
    NotSupportedError
)


DEFAULT_ENGINE = AsyncEngine(retry_handler=AdvancedRetryHandler())


__all__ = [
    'Engine', 'AsyncEngine', 'RetryDirective', 'RetryHandler', 'NoRetryHandler',
    'SimpleRetryHandler', 'LimitedRetryHandler', 'AdvancedRetryHandler', 'PushExecutor',
    'NotSupportedError', 'DEFAULT_ENGINE'
]



"""Basic stuff for plugins (pull, push, udf)."""
import inspect
import logging
from abc import ABCMeta
from importlib import import_module
from typing import Any, Tuple, Optional, Union, cast, Callable, Iterable

from pnp import validator
from pnp.utils import ReprMixin


class InstallOptionalExtraError(ImportError):
    """Is raised when an optional extra is not installed, but a plugin requires it."""
    def __init__(self, extra_name: str):
        super().__init__("You have to install extra '{}' to use this plugin".format(extra_name))


def load_optional_module(namespace: str, extra: str) -> Any:
    """Load an optional module. If the module does not exist an `InstallOptionalExtraError` is
    raised."""
    try:
        return import_module(namespace)
    except ImportError:
        raise InstallOptionalExtraError(extra) from None


class PluginStoppedError(RuntimeError):
    """Is raised when a operation is canceled / refused because the plugin has stopped."""


class PluginLogAdapter(logging.LoggerAdapter):
    """Logging adapter for plugin classes. Will put the name of the plugin in front of the
    message."""
    def __init__(self, logger: logging.Logger, prefix: str):
        super().__init__(logger, {})
        self.prefix = prefix

    def process(self, msg: str, kwargs: Any) -> Tuple[str, Any]:
        return '[{prefix}] {message}'.format(prefix=self.prefix, message=msg), kwargs


class Plugin(ReprMixin, metaclass=ABCMeta):
    """Base class for a plugin."""

    __REPR_FIELDS__ = ['name']

    def __init__(self, name: str, base_path: Optional[str] = None, **kwargs: Any):
        """
        Initializer.

        Args:
            name (str): Name of the plugin.
            base_path (str): Base path the plugin should use.
            **kwargs: Additional arguments.
        """
        super().__init__()
        self.name = str(name)
        self._base_path = base_path and str(base_path)

        for arg_name in kwargs:
            self.logger.warning("Argument %s not consumed by plugin", arg_name)

    @property
    def base_path(self) -> str:
        """Return the base path of the plugin."""
        # Base path is whether injected (most probably were the loaded config is located
        if self._base_path is None:
            # ... or the current working directory
            import os
            return os.getcwd()
        return self._base_path

    @property
    def logger(self) -> logging.Logger:
        """Return the logger of this instance."""
        component = "{}.{}".format(type(self).__module__, type(self).__name__)
        logger = logging.getLogger(component)
        return cast(logging.Logger, PluginLogAdapter(logger, self.name))

    def _assert_abstract_compat(self, abstracts: Iterable[type]) -> None:
        if not isinstance(self, tuple(abstracts)):
            raise TypeError(
                f"Instance is not a valid plugin subclass. Need to subclass one of {abstracts}"
            )

    def _assert_fun_compat(self, fun_name: str) -> None:
        pull_fun = getattr(self, fun_name, None)
        if not pull_fun:
            raise AttributeError(
                f"Instance needs to implement '{fun_name}' method to be a compatible subclass"
            )
        if not inspect.ismethod(pull_fun):
            raise ValueError(
                f"Attribute {fun_name} of this instance is expected to be an instance method, "
                f"but is {type(pull_fun)}",
            )


class ClassNotFoundError(Exception):
    """Is raised when the specified class of a plugin is not found."""


class NamespaceNotFoundError(Exception):
    """Is raised when the namespace of a class is not found."""


class InvocationError(Exception):
    """Is raised when the invocation (creation) of the plugin failed."""


class PluginTypeError(Exception):
    """Is raised when the plugin does not meet the requested plugin type or is no plugin at all."""


def load_plugin(plugin_path: str, plugin_type: Union[type, str], instantiate: bool = True,
                **kwargs: Any) -> Union[Plugin, Callable[..., Any]]:
    """
    Loads a plugin by using a fully qualified identifier (<module_path>.<class_name>,
    e.g. pnp.plugins.pull.simple.Repeat).

    Args:
        plugin_path: Fully qualified path to plugin path.
        plugin_type: Base class the plugin has to extend / inherit from.
        instantiate: If True the class will be instantiated by passing the given **kwargs.
            Otherwise the class (not an object) is returned.
        **kwargs: Plugin arguments.

    Returns:
        If everything went fine an instantiated plugin is returned.
        Multiple things can go wrong:
        - Module not found (NamespaceNotFoundError)
        - Class not found (ClassNotFoundError)
        - Instantiation error (InvocationError)
        - Wrong plugin base type (PluginTypeError)
    """
    validator.is_instance(str, plugin_path=plugin_path)
    validator.is_instance(type, str, plugin_type=plugin_type)
    if isinstance(plugin_type, str) and plugin_type != 'callable':
        raise ValueError("When 'plugin_type' is str, the only allowed value is callable")

    k = plugin_path.rfind('.')
    if k > -1:
        # Namespace = everything before the last '.'
        namespace = plugin_path[:k]
        # Class name = everything after the last '.'
        clazz_name = plugin_path[k + 1:]
    else:
        namespace = 'builtins'
        clazz_name = plugin_path

    try:
        loaded_module = import_module(namespace)
        clazz = getattr(loaded_module, clazz_name)

        if isinstance(plugin_type, str) and plugin_type == 'callable':
            if not callable(clazz):
                raise PluginTypeError("The plugin is requested to be a callable, but it is not.")
        else:
            if not issubclass(clazz, cast(type, plugin_type)):
                raise PluginTypeError(
                    "The plugin is requested to inherit from '{}', but it does not."
                    .format(plugin_type)
                )

        return cast(Plugin, clazz(**kwargs)) if instantiate else cast(Callable[..., Any], clazz)
    except AttributeError:
        raise ClassNotFoundError('Class {} was not found in namespace {}'
                                 .format(clazz_name, namespace)) from None
    except ImportError:
        raise NamespaceNotFoundError("Namespace '{}' not found".format(namespace)) from None
    except TypeError as terr:
        raise InvocationError("Invoked constructor from class '{}' in namespace '{}' failed"
                              .format(clazz_name, namespace)) from terr



"""Data model."""
from typing import Dict, List, Union

from pydantic import BaseModel, Field

from pnp.plugins.pull import Pull
from pnp.plugins.push import Push
from pnp.plugins.udf import UserDefinedFunction
from pnp.typing import AnyCallable, SelectorExpression


class PullModel(BaseModel):
    """Model representing a pull configuration."""

    # The actual pull instance of type `Pull`
    instance: Pull

    class Config:
        """Pydantic configuration."""
        arbitrary_types_allowed = True


class PushModel(BaseModel):
    """Model representing a push."""

    # The actual push instance of type `Push`
    instance: Push

    # The given selector expression or None if not passed.
    selector: SelectorExpression = None

    # If true the selector will be computed for each payload item (if applicable)
    unwrap: bool = False

    # A list of push depdencies
    deps: List['PushModel'] = Field(default_factory=list)

    class Config:
        """Pydantic configuration."""
        arbitrary_types_allowed = True


PushModel.update_forward_refs()


class TaskModel(BaseModel):
    """Model representing a task (pull and dependant pushes)."""

    # The name of the task
    name: str

    # The pull model to produce incoming data
    pull: PullModel

    # List of pushes that are triggered when the pull produced some data
    pushes: List[PushModel]

    class Config:
        """Pydantic configuration."""
        arbitrary_types_allowed = True


class APIModel(BaseModel):
    """Model representing the api server configuration."""

    # The port the api should listen for incoming requests
    port: int

    # Enables the /metrics endpoint
    enable_metrics: bool


class UDFModel(BaseModel):
    """Model representing a user-defined function."""

    # The alias name of the udf
    name: str

    # The callable or the UDF to use
    callable: Union[AnyCallable, UserDefinedFunction]

    class Config:
        """Pydantic configuration."""
        arbitrary_types_allowed = True


# Alias type for a set of tasks
TaskSet = Dict[str, TaskModel]



"""Contains log related endpoints."""

import logging

from fastapi import FastAPI, Query, HTTPException
from starlette.responses import JSONResponse

from pnp.api.models import EmptyResponse
from .base import Endpoint


class SetLogLevel(Endpoint):
    """Set log level endpoint. Change the logging level during runtime."""

    async def endpoint(
            self,
            level: str = Query(
                ...,
                title="Logging level",
                description="One of DEBUG, INFO, WARNING, ERROR, CRITICAL",
                regex="^DEBUG|INFO|WARNING|ERROR|CRITICAL$"
            )
    ) -> JSONResponse:
        """Change the logging level of the application during runtime.
        Make sure you pass a valid log level (DEBUG, INFO, WARNING, ERROR, CRITICAL)."""
        level = level.upper()
        level_code = logging.getLevelName(level)
        if not isinstance(level_code, int):
            raise HTTPException(
                status_code=422,
                detail="Given log level is not valid. Make sure to pass one of "
                       "DEBUG, INFO, WARNING, ERROR"
            )
        logging.getLogger().setLevel(level_code)
        return EmptyResponse()

    def attach(self, fastapi: FastAPI) -> None:
        fastapi.post(
            path="/loglevel",
            response_class=JSONResponse,
            responses={
                500: {"description": "Internal Server Error"}
            }
        )(self.endpoint)



"""Contains a version endpoint."""

import sys

from fastapi import FastAPI
from pydantic import BaseModel  # pylint: disable=no-name-in-module

from pnp import __version__ as __pnp_version__
from .base import Endpoint


class VersionResponse(BaseModel):
    """Version response."""

    version: str = __pnp_version__
    python: str = sys.version


class Version(Endpoint):
    """When requested returns the current pnp and python version."""

    async def _endpoint(self) -> VersionResponse:
        """Returns the current pnp and python version."""
        return VersionResponse()

    def attach(self, fastapi: FastAPI) -> None:
        """Attach the endpoint to the serving component."""
        fastapi.get(
            path="/version",
            response_model=VersionResponse
        )(self._endpoint)



"""Contains a catch all route."""

from typing import Iterable, Callable, Awaitable, List, Dict, Union, Any

from fastapi import FastAPI
from pydantic import BaseModel  # pylint: disable=no-name-in-module
from starlette.requests import Request
from typeguard import typechecked

from pnp import validator
from pnp.api.models import HealthResponse
from pnp.utils import HTTP_METHODS
from .base import Endpoint


class CatchAllRequest(BaseModel):
    """Incoming request that is passed as an argument to the callback when
    a request hits the specified route."""
    endpoint: str
    levels: List[str]
    method: str
    query_params: Dict[str, Union[List[Any], Any]]
    data: Any
    is_json_data: bool
    url: str
    full_path: str
    path: str


# Callback when a request hits the route
CatchAllCallback = Callable[[CatchAllRequest], Awaitable[None]]


class CatchAllRoute(Endpoint):
    """Catch all requests under a specific route prefix."""

    @typechecked
    def __init__(
            self, route_prefix: str, allowed_methods: Iterable[str], callback: CatchAllCallback
    ):
        self.route_prefix = str(route_prefix)
        self.allowed_methods = list(allowed_methods)
        self.callback = callback
        validator.subset_of(HTTP_METHODS, allowed_methods=self.allowed_methods)

    async def _endpoint_root(self, request: Request) -> HealthResponse:
        """Any requests made to this endpoint will be tracked by the attached
        pull.http.Server component."""
        return await self._process_request(request, '/')

    async def _endpoint_subpath(self, request: Request, sub_path: str) -> HealthResponse:
        """Any requests made to this endpoint will be tracked by the attached
        pull.http.Server component."""
        return await self._process_request(request, sub_path)

    async def _process_request(self, request: Request, path: str) -> HealthResponse:
        assert self.callback

        try:
            data = await request.json()
            is_json = True
        except ValueError:
            data = (await request.body())
            if data == b"":
                data = None
            is_json = False

        query_params: Dict[str, Any] = {}
        for param_key in request.query_params:
            # print(param_key, value, request.query_params.getlist(key))
            val = request.query_params.getlist(param_key)
            if len(val) == 1:
                query_params[param_key] = val[0]  # Single item list only
            else:
                query_params[param_key] = val

        callback_request = CatchAllRequest(
            endpoint=path,
            levels=["/"] if path == "/" else path.split('/'),
            method=request.method,
            query_params=query_params,
            data=data,
            is_json_data=is_json,
            url=str(request.url),
            full_path="{}?{}".format(request.url.path, request.url.query),
            path=request.url.path
        )

        await self.callback(callback_request)
        return HealthResponse()

    def attach(self, fastapi: FastAPI) -> None:
        """Attach the endpoint to the serving component."""
        for method in self.allowed_methods:
            fun = getattr(fastapi, method.lower(), None)
            if fun:
                fun(
                    path=f"/{self.route_prefix}",
                    response_model=HealthResponse,
                )(self._endpoint_root)
                fun(
                    path=f"/{self.route_prefix}/{{sub_path:path}}",
                    response_model=HealthResponse
                )(self._endpoint_subpath)



"""Utility functions related to zipping and archiving."""

import logging
import os
import zipfile
from typing import Optional, Union, Iterable

import pathspec

from pnp import validator
from pnp.utils import is_iterable_but_no_str


_LOGGER = logging.getLogger(__file__)


def zipdir(
    source_dir: str, target_zip: str, ignore: Optional[Union[str, Iterable[str]]] = None
) -> None:
    """Creates a zip file specified by `out_zip`. The source directory is specified
    by `source`. You can optionally specify a gitignore-like ignore list to ignore
    specific items."""
    source_dir = str(source_dir)
    validator.is_directory(source=source_dir)

    target_zip = str(target_zip)

    if ignore is None:
        ignore = []
    if not is_iterable_but_no_str(ignore):
        ignore = [str(ignore)]
    ignore = [str(item) for item in ignore]

    matcher = pathspec.PathSpec.from_lines('gitwildmatch', ignore)
    source_dir = os.path.abspath(source_dir)  # Make relative path -> absolute
    if not source_dir.endswith(os.sep):
        source_dir += os.sep

    _LOGGER.debug("Creating '%s' from '%s'", target_zip, source_dir)
    with zipfile.ZipFile(target_zip, 'w', zipfile.ZIP_DEFLATED) as zipper:
        for root, _, files in os.walk(source_dir):
            for file in files:
                fpath = os.path.join(root, file)
                arcname = fpath.replace(source_dir, '')
                if not matcher.match_file(fpath):  # Apply ignore patterns
                    _LOGGER.debug("Zipping '%s' as '%s", fpath, arcname)
                    zipper.write(fpath, arcname=arcname)
                else:
                    _LOGGER.debug("Ignored '%s'", fpath)


def zipfiles(source_files: Optional[Union[str, Iterable[str]]], target_zip: str) -> None:
    """Zips one or multiple files into a zip file. The target zip file is specified by
    `out_zip`."""
    if not source_files:
        return
    if not is_iterable_but_no_str(source_files):
        source_files = [str(source_files)]
    source_files = [str(fitem) for fitem in source_files]

    _LOGGER.debug("Creating '%s' from %s files", target_zip, len(source_files))
    with zipfile.ZipFile(target_zip, 'w', zipfile.ZIP_DEFLATED) as zipper:
        for fpath in source_files:
            arcname = os.path.basename(fpath)
            _LOGGER.debug("Zipping '%s' as '%s'", fpath, arcname)
            zipper.write(fpath, arcname)


def zipignore(path: str, file_name: str = '.zipignore') -> Iterable[str]:
    """Parses a zipignore file."""
    # Check: Is valid directory
    def filter_(line: str) -> str:
        line = line.strip()  # Remove whitespace characters
        if line.startswith('#'):  # Comment
            return ''  # Will be filtered later
        return line

    zipignore_path = os.path.join(path, file_name)
    if not os.path.isfile(zipignore_path):
        return []  # Shortcircuit -> no file -> no rules

    with open(zipignore_path, 'r') as zih:
        lines = zih.readlines()
    lines.append(file_name)  # Ignore zipignore file name as well

    return filter(
        lambda item: item != '',
        [filter_(line) for line in lines]
    )



"""Sound related shared stuff."""

import logging
import os
from typing import Any, Tuple, Optional

from typeguard import typechecked

from pnp import validator

# Cannot import numpy because it's an extra package and might not be available at runtime
NumpyNDArray = Any


_LOGGER = logging.getLogger(__file__)


MODE_PEARSON = 'pearson'
MODE_STD = 'std'
ALLOWED_MODES = [MODE_PEARSON, MODE_STD]


class WavFile:
    """Wav file representation including. FFT representation."""
    @typechecked
    def __init__(self, abs_path: str):
        self.abs_path = abs_path
        self._file_name: Optional[str] = None
        self._signal_fft: Optional[NumpyNDArray] = None
        self._signal_length: Optional[int] = None

    @property
    def file_name(self) -> str:
        """Return the name of the wav file (without path)."""
        return os.path.basename(os.path.splitext(self.abs_path)[0])

    @property
    def signal_fft(self) -> NumpyNDArray:
        """Return the fft representation of this wav file."""
        if self._signal_fft is None:
            self._signal_length, self._signal_fft = self._load_wav_fft()
        return self._signal_fft

    @property
    def signal_length(self) -> int:
        """Return the signal length of this wav file."""
        if not self._signal_length:
            self._signal_length, self._signal_fft = self._load_wav_fft()
        return self._signal_length

    def _load_wav_fft(self) -> Tuple[int, NumpyNDArray]:
        import scipy.io.wavfile as wavfile
        _LOGGER.debug("Loading wav file from '%s'", self.abs_path)
        sample_rate, signal = wavfile.read(self.abs_path)
        N, secs, signal_fft = perform_fft(signal, sample_rate)
        _LOGGER.debug("Loaded %s seconds wav file @ %s hz", secs, sample_rate)
        return N, signal_fft

    @classmethod
    def from_path(cls, wav_file: str, base_path: str) -> 'WavFile':
        """Load the wav file from a relative or absolute path."""
        def _mk_abs(file_item: str) -> str:
            file_item = str(file_item)
            if os.path.isabs(file_item):
                return file_item
            return os.path.join(base_path, file_item)

        wav_file = str(wav_file)
        abs_path = _mk_abs(wav_file)
        validator.is_file(wav_file=abs_path)

        return WavFile(abs_path)


def perform_fft(
    signal: NumpyNDArray, rate: int, add_zeros: bool = True
) -> Tuple[int, float, NumpyNDArray]:
    """Perform a fast fournier transformation on a given signal."""
    import numpy as np
    try:
        import scipy.fft
        fft_fun = scipy.fft.fft  # pylint: disable=no-member
    except ImportError:
        import scipy
        fft_fun = scipy.fft  # pylint: disable=no-member

    chn = len(signal.shape)
    if chn >= 2:  # Make mono channel
        signal = signal.sum(axis=1) / 2
    N = int(signal.shape[0])
    secs = N / float(rate)
    if add_zeros:
        signal = np.concatenate((signal, np.zeros(len(signal))), axis=None)
    trans_fft = abs(fft_fun(signal))

    return N, secs, trans_fft


def similarity_pearson(
    buffer: NumpyNDArray, wav_file: WavFile, rate: int, threshold: float
) -> Tuple[bool, float, float]:
    """Check the similarity of a recorded sound buffer and a given wav_file.
    Use the pearson correlation."""
    _, _, current_fft = perform_fft(buffer[:wav_file.signal_length], rate, True)

    from scipy.stats.stats import pearsonr
    corrcoef = pearsonr(wav_file.signal_fft, current_fft)[0]

    return corrcoef >= threshold, corrcoef, threshold


def similarity_std(
        buffer: NumpyNDArray, wav_file: WavFile, rate: int, threshold: float
) -> Tuple[bool, float, float]:
    """Check the similarity of a recorded sound buffer and a given wav_file.
    Use a correlation check using the standard deviation."""
    import numpy as np
    _, _, current_fft = perform_fft(buffer[:wav_file.signal_length], rate, True)
    corrcoef = np.correlate(
        wav_file.signal_fft / wav_file.signal_fft.std(),
        current_fft / current_fft.std()
    )[0] / len(wav_file.signal_fft)

    return corrcoef >= threshold, corrcoef, threshold



"""Machine learning related push plugins."""

import io
import os

from pnp import validator
from pnp.plugins import load_optional_module
from pnp.plugins.push import SyncPush, enveloped, drop_envelope
from pnp.utils import make_list
from pnp.validator import one_not_none


class FaceR(SyncPush):
    """
    FaceR (short one for face recognition) tags known faces in images.
    Output is the image with all faces tagged whether with the known name or an `unknown_label`.
    Default for unknown ones is 'Unknown'.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/push/ml.FaceR/index.md
    """
    __REPR_FIELDS__ = ['known_faces', 'known_faces_dir', 'known_names']

    EXTRA = 'faceR'

    def __init__(self, known_faces=None, known_faces_dir=None, unknown_label="Unknown", lazy=False,
                 **kwargs):
        # known_faces -> mapping name -> list of files
        # known_faces_dir -> directory with known faces (filename -> name)
        # unknown_label -> Label for unknown faces
        super().__init__(**kwargs)

        one_not_none(known_faces=known_faces, known_faces_dir=known_faces_dir)
        self.known_faces = known_faces
        validator.is_instance(dict, allow_none=True, known_faces=self.known_faces)
        self.known_faces_dir = known_faces_dir and str(known_faces_dir)
        if self.known_faces_dir:
            validator.is_directory(known_faces_dir=self.known_faces_dir)

        if not lazy:
            self._configure()
        else:
            self.known_names = None
            self.known_encodings = None
            self.face_recognition = None

        self.unknown_label = str(unknown_label)

    def _load_fencodings(self, fp):
        img = self.face_recognition.load_image_file(fp)
        return self.face_recognition.face_encodings(img)[0]

    def _load_from_mapping(self, mapping):
        def _loop():
            for name, fps in mapping.items():
                for fp in make_list(fps):
                    yield name, self._load_fencodings(fp)
        targets = list(_loop())
        return [name for name, _ in targets], [fp for _, fp in targets]

    def _load_from_directory(self, directory):
        def _loop():
            accepted = ('.tif', '.tiff', '.gif', '.jpeg', '.jpg', '.jif', '.jfif', '.png')
            for file in os.listdir(os.fsencode(directory)):
                filename = os.fsdecode(file)
                if filename.lower().endswith(accepted):
                    # strip the extension from the filename
                    yield (
                        os.path.splitext(filename)[0],
                        self._load_fencodings(os.path.join(directory, filename))
                    )
        targets = list(_loop())
        return [name for name, _ in targets], [fp for _, fp in targets]

    def _configure(self):
        # Do not break the complete module, when extra_packages are not present
        self.face_recognition = load_optional_module('face_recognition', self.EXTRA)

        # If both are set known_faces is the default
        if self.known_faces:
            self.known_names, self.known_encodings = self._load_from_mapping(self.known_faces)
        else:
            self.known_names, self.known_encodings = self._load_from_directory(self.known_faces_dir)

    @staticmethod
    def _tag_image(unknown_image, tags):
        from PIL import Image, ImageDraw  # pylint: disable=import-error

        # Convert the image to a PIL-format image so that we can draw on top of it with the
        # Pillow library. See http://pillow.readthedocs.io/ for more about PIL/Pillow
        pil_image = Image.fromarray(unknown_image)
        # Create a Pillow ImageDraw Draw instance to draw with
        draw = ImageDraw.Draw(pil_image)
        try:
            for name, (top, right, bottom, left) in tags:
                # Draw a box around the face using the Pillow module
                draw.rectangle(((left, top), (right, bottom)), outline=(0, 0, 255))

                # Draw a label with a name below the face
                _, text_height = draw.textsize(name)
                draw.rectangle(
                    ((left, bottom - text_height - 10), (right, bottom)),
                    fill=(0, 0, 255),
                    outline=(0, 0, 255)
                )
                draw.text((left + 6, bottom - text_height - 5), name, fill=(255, 255, 255, 255))

            return pil_image
        finally:
            # Remove the drawing library from memory as per the Pillow docs
            del draw

    @enveloped
    @drop_envelope
    def _push(self, payload):
        if not self.face_recognition:
            self._configure()

        # Load unknown image and find faces
        unknown_image = self.face_recognition.load_image_file(io.BytesIO(payload))
        face_locations = self.face_recognition.face_locations(unknown_image)
        face_encodings = self.face_recognition.face_encodings(unknown_image, face_locations)

        tags = []
        # Loop through each face found in the unknown image
        for flocation, face_encoding in zip(face_locations, face_encodings):
            # See if the face is a match for the known face(s)
            matches = self.face_recognition.compare_faces(self.known_encodings, face_encoding)
            name = self.unknown_label
            # If a match was found in known_face_encodings, just use the first one.
            if True in matches:
                first_match_index = matches.index(True)
                name = self.known_names[first_match_index]
            tags.append((name, flocation))

        # Tag the face locations
        tagged_image = self._tag_image(unknown_image, tags)

        # Save the image to a bytestream
        out_io = io.BytesIO()
        try:
            tagged_image.save(out_io, format='PNG')
            return dict(
                tagged_image=out_io.getvalue(),
                no_of_faces=len(face_locations),
                known_faces=list({name for name, _ in tags if name != self.unknown_label})
            )
        finally:
            del out_io



"""Contains file system related push plugins."""

import os
import time

from pnp import validator
from pnp.plugins.push import SyncPush, enveloped, parse_envelope, drop_envelope


class FileDump(SyncPush):
    """
    This push dumps the given `payload` to a file to the specified `directory`.
    If argument `file_name` is None, a name will be generated based on the current datetime
    (%Y%m%d-%H%M%S).
    If `file_name` is not passed (or None) you should pass `extension` to specify the extension of
    the generated file name.
    Argument `binary_mode` controls whether the dump is binary (mode=wb) or text (mode=w).


    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/push/fs.FileDump/index.md

    Examples:

        >>> import tempfile
        >>> import asyncio
        >>> loop = asyncio.get_event_loop()
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     # Automatic generated file name with extension '.dump'
        ...     dut = FileDump(name='doctest', directory=tmpdir, binary_mode=False)
        ...     created_file = loop.run_until_complete(dut.push("I am the content"))
        ...     with open(created_file, 'r') as fs:
        ...         assert fs.read() == "I am the content"

    """
    __REPR_FIELDS__ = ['binary_mode', 'directory', 'extension', 'file_name']

    def __init__(self, directory='.', file_name=None, extension='.dump', binary_mode=True,
                 **kwargs):
        super().__init__(**kwargs)
        self.directory = directory
        validator.is_directory(directory=self.directory)
        self.extension = self._parse_extension(extension)
        self.binary_mode = bool(binary_mode)
        self.file_name = self._parse_file_name(file_name)

    @staticmethod
    def _parse_file_name(value):
        return value and str(value)

    @staticmethod
    def _parse_extension(value):
        value = str(value)
        if not value.startswith('.'):
            return'.' + value
        return value

    @enveloped
    @parse_envelope('file_name')
    @parse_envelope('extension')
    @drop_envelope
    def _push(self, file_name, extension, payload):  # pylint: disable=arguments-differ
        if file_name is None:
            file_name = time.strftime("%Y%m%d-%H%M%S")
        file_path = os.path.join(self.directory, file_name + extension)
        with open(file_path, 'wb' if self.binary_mode else 'w') as fhandle:
            fhandle.write(str(payload) if not self.binary_mode else payload)

        return file_path


class Zipper(SyncPush):
    """
    Zips the given source directory or file and returns the path to the created zip
    archive.

    Args:
        source (str): Specifies the source directory or file to zip. If not passed the
            source can be specified by the envelope at runtime.
        out_path (str): Specifies the path to the general output path where all target
            zip files should be generated. If not passed the systems temp directory is
            used.

    Returns:
        Returns the absolute path to the created zip file.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/push/fs.Zipper/index.md

    """
    __REPR_FIELDS__ = ['archive_name', 'out_path', 'source']

    def __init__(self, source=None, out_path=None, archive_name=None, **kwargs):
        super().__init__(**kwargs)
        self.source = self._parse_source(source)
        self.out_path = self._parse_out_path(out_path)
        self.archive_name = self._parse_archive_name(archive_name)

    @staticmethod
    def _parse_source(value):
        if not value:
            return None
        value = str(value)
        if not os.path.isfile(value) and not os.path.isdir(value):
            raise ValueError("Argument 'source' is neither a directory nor a file")
        return value

    @staticmethod
    def _parse_out_path(value):
        if not value:
            import tempfile
            return tempfile.gettempdir()
        validator.is_directory(out_path=value)
        return os.path.abspath(str(value))

    @staticmethod
    def _parse_archive_name(value):
        return value and str(value)

    @enveloped
    @parse_envelope('archive_name')
    @drop_envelope
    def _push(self, archive_name, payload):  # pylint: disable=arguments-differ
        source = self._parse_source(self.source or payload)

        if os.path.isdir(source):
            # Directory branch
            from ...shared.zipping import zipdir, zipignore
            base_dir_name = os.path.basename(os.path.normpath(source))
            out_file = (
                os.path.join(self.out_path, archive_name) if archive_name else
                os.path.join(self.out_path, base_dir_name + '.zip')
            )
            ignore_list = zipignore(source)
            zipdir(source, out_file, ignore_list)
            return out_file

        # File branch
        from ...shared.zipping import zipfiles
        out_file = (
            os.path.join(self.out_path, archive_name) if archive_name else
            os.path.join(self.out_path, os.path.basename(source) + '.zip')
        )
        zipfiles(source, out_file)
        return out_file



"""Asyncio helper."""

import asyncio
from typing import Callable, Any, Coroutine

from pnp import validator
from pnp.typing import T
from pnp.utils import StopCycleError

SleepInterruptPredicate = Callable[[], Coroutine[Any, Any, bool]]
SleepInterruptCallback = Callable[[], Coroutine[Any, Any, None]]


async def async_interruptible_sleep(
    wait: float, callback: SleepInterruptCallback, interval: float = 0.1
) -> None:
    """
    Waits the specified amount of time. The waiting can be interrupted when the callback raises a
    `StopCycleError`. The argument `interval` defines after how much waiting time the callback
    should be called to determine if the sleep should be interrupted or not.
    """
    wait = float(wait)
    interval = float(interval)

    complete_cycles = int(wait // interval)
    try:
        for _ in range(0, complete_cycles):
            await callback()  # Should raise a StopCycleError error when waiting should be aborted
            await asyncio.sleep(interval)

        await asyncio.sleep(wait % interval)
    except StopCycleError:
        pass


async def async_sleep_until_interrupt(
    sleep_time: float, interrupt_fun: SleepInterruptPredicate, interval: float = 0.1
) -> None:
    """Call this method to sleep an interruptable sleep until the interrupt co-routine returns
    True."""
    validator.is_function(interrupt_fun=interrupt_fun)

    async def callback() -> None:
        if await interrupt_fun():
            raise StopCycleError()
    await async_interruptible_sleep(sleep_time, callback, interval=interval)


async def run_sync(func: Callable[..., T], *args: Any) -> T:
    """Runs sync code in an async compatible non-blocking way using an executor."""
    loop = asyncio.get_event_loop()
    return await loop.run_in_executor(None, func, *args)



"""Shared Exceptions."""


class TemplateError(Exception):
    """Is raised when a template cannot be rendered."""


class NoEngineError(Exception):
    """Is raised when no engine is supplied but is needed."""



"""Contains simple user-defined functions."""

from pnp.plugins.udf import UserDefinedFunction


class Counter(UserDefinedFunction):
    """
    Memories a counter value which is increased everytime you call the udf.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/udf/simple.Counter/index.md

    Example:

        >>> dut = Counter(name='doctest')
        >>> dut(), dut(), dut()
        (0, 1, 2)
        >>> dut = Counter(init=5, name='doctest')
        >>> dut(), dut(), dut()
        (5, 6, 7)
        >>> dut = Counter(init=-5, name='doctest')
        >>> dut(), dut(), dut()
        (0, 1, 2)
        >>> dut = Counter(init="5", name='doctest')
        >>> dut(), dut(), dut()
        (5, 6, 7)
    """
    __REPR_FIELDS__ = ['cnt']

    def __init__(self, init=0, **kwargs):
        super().__init__(**kwargs)
        self.cnt = int(init)
        if self.cnt < 0:
            self.cnt = 0

    def action(self):  # pylint: disable=arguments-differ
        try:
            return self.cnt
        finally:
            self.cnt += 1


class FormatSize(UserDefinedFunction):
    """
    Returns the size of a file (or whatever) as a human readable size (e.g. bytes, KB, MB, GB, TB,
    PB). The input is expected to be at byte scale.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/udf/simple.FormatSize/index.md

    Example:
        >>> dut = FormatSize(name='doctest')
        >>> dut.action(0)
        '0 byte'
        >>> dut.action(1)
        '1 byte'
        >>> dut.action(5555.5)
        '5 KB'
        >>> dut.action(4672512)
        '4.5 MB'
        >>> dut.action(45548128174.08)
        '42.42 GB'
        >>> dut.action(-42)
        Traceback (most recent call last):
        ...
        ValueError: No negative values supported.
    """
    @staticmethod
    def _human_size(size_bytes):
        """
        format a size in bytes into a 'human' file size, e.g. bytes, KB, MB, GB, TB, PB
        Note that bytes/KB will be reported in whole numbers but MB and above will have
        greater precision e.g. 1 byte, 43 bytes, 443 KB, 4.3 MB, 4.43 GB, etc

        Taken from:
        https://stackoverflow.com/questions/1094841/reusable-library-to-get-human-readable-version-of-file-size
        """
        num = float(size_bytes)
        if num < 0:
            raise ValueError("No negative values supported.")

        if num <= 1:
            # because I really hate unnecessary plurals
            return "%d byte" % num

        suffixes_table = [('bytes', 0), ('KB', 0), ('MB', 1), ('GB', 2), ('TB', 2), ('PB', 2)]

        # Will at least run once
        for suffix, precision in suffixes_table:
            if num < 1024.0:
                break
            num /= 1024.0

        if precision == 0:
            formatted_size = "%d" % num
        else:
            formatted_size = str(round(num, ndigits=precision))

        return "%s %s" % (formatted_size, suffix)

    def action(self, size_in_bytes):  # pylint: disable=arguments-differ
        return self._human_size(size_in_bytes)


class Memory(UserDefinedFunction):
    """
    Returns a previously memorized value when called.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/udf/simple.Memory/index.md

    Example:

        >>> dut = Memory(name='doctest', init='foo')
        >>> dut(), dut(new_memory='baz'), dut()
        ('foo', 'foo', 'baz')
    """
    __REPR_FIELDS__ = ['current']

    _MISSING = object()

    def __init__(self, init=_MISSING, **kwargs):
        super().__init__(**kwargs)
        self.current = init

    def action(self, new_memory=_MISSING):  # pylint: disable=arguments-differ
        try:
            if self.current is self._MISSING:
                return None if new_memory is self._MISSING else new_memory

            return self.current
        finally:
            if new_memory is not self._MISSING:
                self.current = new_memory



"""Network related plugins."""

import socket
import ssl
from datetime import datetime, timedelta
from typing import Tuple

from glom import glom

from pnp.plugins import load_optional_module
from pnp.plugins.pull import SyncPolling
from pnp.typing import Payload
from pnp.utils import is_local, bps_mbps


class PortProbe(SyncPolling):
    """
    Periodically establishes socket connection to check if anybody is listening
    on a given server on a specific port.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/pull/net.PortProbe/index.md
    """
    __REPR_FIELDS__ = ['port', 'server', 'timeout']

    CONST_SERVER = 'server'
    CONST_PORT = 'port'
    CONST_REACHABLE = 'reachable'

    def __init__(self, port, server='localhost', timeout=1, **kwargs):
        super().__init__(**kwargs)
        self.port = int(port)
        self.server = str(server)
        self.timeout = int(timeout)

    def _poll(self):
        if is_local(self.server):
            available = self._local_probe()
        else:
            available = self._remote_probe()

        return {
            self.CONST_SERVER: self.server,
            self.CONST_PORT: self.port,
            self.CONST_REACHABLE: available
        }

    def _remote_probe(self):
        """Will open a socket connection and try to connect to the specified port."""
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            sock.settimeout(self.timeout)
            try:
                result = sock.connect_ex((self.server, self.port))
                # This might fail if the socket couldn't connect
                sock.shutdown(socket.SHUT_RDWR)
            except socket.gaierror:
                result = 1
            except OSError:
                pass
            return result == 0

    def _local_probe(self):
        """Will try to run a socket server listening on the specified port. This is a
        much better approach compared to try to connect to a server. Connecting to the
        server might flood logs, trigger actions, ..."""
        try:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
                sock.bind(('127.0.0.1', self.port))
                sock.listen(0)
            with socket.socket(socket.AF_INET6, socket.SOCK_STREAM) as sock:
                sock.bind(('::1', self.port))
                sock.listen(0)
            return False
        except socket.error:
            return True


class Speedtest(SyncPolling):
    """
    Periodically tests your real internet download / upload speed as well as the ping latency.
    """

    EXTRA = 'speedtest'

    OUTPUT_SPEC = {
        'download_speed_bps': 'download',  # in Bits/s
        'download_speed_mbps': ('download', bps_mbps),  # in Megabits/s
        'upload_speed_bps': 'upload',  # in Bits/s
        'upload_speed_mbps': ('upload', bps_mbps),  # in Bits/s
        'ping_latency': 'ping',  # in ms
        'result_image': 'share',  # Image provided by speedtest.net to share
        'server': {
            'name': 'server.sponsor',  # The sponsor of the host
            'host': 'server.host',  # The host name
            'location': {
                'city': 'server.name',
                'country': 'server.country',
                'lat': 'server.lat',
                'lon': 'server.lon'
            }
        },
        'client': {
            'isp': 'client.isp',  # Your Internet service provider
            'rating': 'client.isprating'  # The rating of your isp
        }
    }

    def _poll(self) -> Payload:
        speedtest = load_optional_module('speedtest', self.EXTRA)

        client = speedtest.Speedtest()
        client.get_servers()
        client.get_best_server()
        client.download(threads=None)
        client.upload(threads=None)
        client.results.share()

        res = client.results.dict()
        return glom(res, self.OUTPUT_SPEC)


class SSLVerify(SyncPolling):
    """
    Periodically checks if the ssl certificate of a given host is valid and how
    many days are remaining before the certificate will expire.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/pull/net.SSLVerify/index.md
    """
    __REPR_FIELDS__ = ['host', 'timeout']

    def __init__(self, host: str, timeout: float = 3.0, **kwargs):
        super().__init__(**kwargs)
        self.host = str(host)
        self.timeout = float(timeout)

    def _ssl_expiry_datetime(self) -> Tuple[datetime, timedelta]:
        """
        Fetches the expiration date of the ssl certificate.

        Returns:
            Tuple that contains the timestamp the certificate will expire
            and the number of days left.
        """
        ssl_date_fmt = r'%b %d %H:%M:%S %Y %Z'

        context = ssl.create_default_context()
        conn = context.wrap_socket(
            socket.socket(socket.AF_INET),
            server_hostname=self.host,
        )
        conn.settimeout(self.timeout)

        conn.connect((self.host, 443))
        self.logger.debug("SSL Host: %s", self.host)
        ssl_info = conn.getpeercert()
        self.logger.debug("Connection to host %s successful", self.host)

        # parse the string from the certificate into a Python datetime object
        expires = datetime.strptime(ssl_info['notAfter'], ssl_date_fmt)
        self.logger.debug('SSL certificate for %s expires at %s', self.host, expires.isoformat())

        return expires, expires - datetime.utcnow()

    def _poll(self) -> Payload:
        expires_at, expires_delta = self._ssl_expiry_datetime()
        return {
            'host': self.host,
            'payload': {
                'expires_days': expires_delta.days,
                'expires_at': expires_at,
                'expired': expires_delta <= timedelta(days=0)
            }
        }



"""Presence / Occupancy related plugins."""

from typing import Dict, Any, List, Optional

from pnp import validator
from pnp.plugins import load_optional_module
from pnp.plugins.pull import SyncPolling
from pnp.typing import Payload


class FritzBoxTracker(SyncPolling):
    """Tracks connected / not connected devices using the Fritz!Box api."""

    __REPR_FIELDS__ = ['host', 'offline_delay', 'user', 'whitelist']

    EXTRA = 'fritz'

    CONF_DEFAULT_IP = '169.254.1.1'  # This IP should be valid for all FRITZ!Box routers.
    CONF_DEFAULT_USER = 'admin'
    CONF_DEFAULT_PASSWORD = ''

    def __init__(
            self,
            host: str = CONF_DEFAULT_IP, user: str = CONF_DEFAULT_USER,
            password: str = CONF_DEFAULT_PASSWORD, offline_delay: int = 0,
            whitelist: Optional[List[str]] = None, **kwargs
    ):
        super().__init__(**kwargs)
        self.user = str(user)
        self.password = str(password)
        self.host = str(host)
        self.offline_delay = int(offline_delay)
        if self.offline_delay < 0:
            self.offline_delay = 0
        if whitelist:
            validator.is_iterable_but_no_str(whitelist=whitelist)
        self.whitelist = whitelist and [str(item) for item in whitelist]
        self.fritz_box = None
        self._cache = {}

    def _setup(self):
        if self.fritz_box:
            # Nothing to do
            return

        fritz_hosts = load_optional_module('fritzconnection.lib.fritzhosts', self.EXTRA)
        try:
            self.fritz_box = fritz_hosts.FritzHosts(
                address=self.host,
                user=self.user,
                password=self.password
            )
            if not self.fritz_box.modelname:
                raise ValueError()
        except Exception:  # pylint: disable=broad-except
            self.logger.exception(
                "Cannot connect to the Fritz!Box @ %s. Please review the configuration",
                self.host
            )
            self.fritz_box = None

        self.logger.info("Connected to the Fritz!Box @ %s", self.host)

    def _parse_host(self, host_entry: Dict[str, Any]):
        _ = self  # Fake usage
        if 'NewIPAddress' in host_entry:  # Result structure from specific host call (whitelist)
            return {
                'mac': host_entry['mac'],
                'ip': host_entry['NewIPAddress'],
                'status': host_entry['NewActive'],
                'name': host_entry['NewHostName']
            }
        return host_entry

    def _adjust_status(self, device: Dict[str, Any]) -> Dict[str, Any]:
        mac, status = device['mac'], device['status']
        if status:  # Online, no adjustment
            self._cache[mac] = 0
            return device

        # Status = offline
        if mac not in self._cache:
            self._cache[mac] = self.offline_delay
            return device  # Never seen before -> no adjustment

        if self._cache[mac] < self.offline_delay:
            self._cache[mac] += 1
            device['status'] = True  # Fake online

        return device

    def _get_host_list(self):
        if not self.whitelist:
            return self.fritz_box.get_hosts_info()  # Get all hosts

        # Whitelist is used: Different call to fritzbox
        res = []
        for mac in self.whitelist:
            try:
                device = self.fritz_box.get_specific_host_entry(mac_address=str(mac))
                device['mac'] = str(mac)
                res.append(device)
            except KeyError:
                self.logger.warning("Fritz!Box does not know about the device '%s'", str(mac))

        return res

    def _poll(self) -> Payload:
        self._setup()
        if not self.fritz_box:
            return None  # No connection, no data

        return [
            self._adjust_status(self._parse_host(host))
            for host in self._get_host_list()
        ]



"""MQTT related push plugins."""

from dictmentor import DictMentor, ext

from pnp import validator
from pnp.plugins.push import SyncPush, enveloped, parse_envelope, drop_envelope
from pnp.shared.mqtt import MQTTBase
from pnp.utils import try_parse_bool


class Discovery(MQTTBase, SyncPush):
    """
    See Also:
        https://pnp.readthedocs.io/en/stable/plugins/index.html#mqtt-discovery
    """
    __REPR_FIELDS__ = ['component', 'discovery_prefix', 'node_id', 'object_id']

    SUPPORTED_COMPONENTS = ['alarm_control_panel', 'binary_sensor', 'camera', 'cover', 'fan',
                            'climate', 'light', 'lock', 'sensor', 'switch']

    CONST_STATE_TOPIC = 'state_topic'
    CONST_JSON_ATTRIBUTES_TOPIC = 'json_attributes_topic'

    def __init__(self, discovery_prefix, component, config, object_id=None, node_id=None, **kwargs):
        super().__init__(**kwargs)
        self.discovery_prefix = str(discovery_prefix)
        validator.one_of(self.SUPPORTED_COMPONENTS, component=str(component))
        self.component = str(component)
        self.object_id = self._parse_object_id(object_id)
        validator.is_instance(dict, config=config)
        self._config = config
        self.node_id = self._parse_node_id(node_id)
        self.configured = {}

    @property
    def config(self):
        """Return the mqtt discovery configuration."""
        import copy
        return copy.copy(self._config)

    @staticmethod
    def _parse_object_id(val):
        return str(val)

    @staticmethod
    def _parse_node_id(val):
        return val and str(val)

    def _topics(self, object_id, node_id):  # pylint: disable=unused-argument
        node_id = "{node_id}/".format(node_id=node_id) if node_id else ""
        base_topic = "{prefix}/{component}/{node_id}{object_id}".format(
            prefix=self.discovery_prefix,
            component=self.component,
            node_id=node_id,
            object_id=object_id
        )
        state_topic = base_topic + "/state"
        config_topic = base_topic + "/config"
        attr_topic = base_topic + "/attributes"
        return base_topic, config_topic, state_topic, attr_topic

    def _configure(self, object_id, node_id):
        configure_key = str(object_id) + str(node_id)
        if configure_key not in self.configured:
            base_topic, config_topic, state_topic, attr_topic = self._topics(object_id, node_id)
            mentor = DictMentor(
                ext.Variables(
                    fail_on_unset=True,
                    discovery_prefix=self.discovery_prefix,
                    component=self.component,
                    object_id=object_id,
                    node_id=node_id,
                    base_topic=base_topic,
                    config_topic=config_topic,
                    **{
                        self.CONST_STATE_TOPIC: state_topic,
                        self.CONST_JSON_ATTRIBUTES_TOPIC: attr_topic
                    }
                )
            )
            config_augmented = mentor.augment(self.config)
            if self.CONST_STATE_TOPIC in config_augmented:
                self.logger.warning(
                    "%s is part of your config, but will be ignored",
                    self.CONST_STATE_TOPIC
                )
            if self.CONST_JSON_ATTRIBUTES_TOPIC in config_augmented:
                self.logger.warning(
                    "%s is part of your config, but will be ignored",
                    self.CONST_JSON_ATTRIBUTES_TOPIC
                )

            config_augmented[self.CONST_STATE_TOPIC] = state_topic
            config_augmented[self.CONST_JSON_ATTRIBUTES_TOPIC] = attr_topic

            self._publish(config_augmented, config_topic, retain=True)
            self.configured[configure_key] = True

    @enveloped
    @parse_envelope('object_id')
    @parse_envelope('node_id')
    @parse_envelope('attributes')
    @drop_envelope
    def _push(self, object_id, node_id, attributes, payload):  # pylint: disable=arguments-differ
        if object_id is None:
            raise ValueError("object_id was not defined either by the __init__ nor by the envelope")

        self._configure(object_id, node_id)
        _, _, state_topic, attribute_topic = self._topics(object_id, node_id)
        self._publish(payload, state_topic, retain=True)

        if attributes and isinstance(attributes, dict):
            self._publish(attributes, attribute_topic, retain=True)

        return payload


class Publish(MQTTBase, SyncPush):
    """
    This push will push the given `payload` to a mqtt broker (in this case mosquitto).
    The broker is specified by `host` and `port`. In addition a topic needs to be specified
    were the payload is pushed to (e.g. home/living/thermostat).

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/push/mqtt.Publish/index.md
    """
    __REPR_FIELDS__ = ['multi', 'retain', 'topic']

    def __init__(self, topic=None, retain=False, multi=False, **kwargs):
        super().__init__(**kwargs)
        self.topic = self._parse_topic(topic)
        self.retain = self._parse_retain(retain)
        self.multi = bool(multi)

    @staticmethod
    def _parse_retain(val):
        return try_parse_bool(val, default=False)

    @staticmethod
    def _parse_topic(val):
        return str(val) if val is not None else None

    @staticmethod
    def _topic_concat(topic1, topic2):
        if str(topic1).endswith('/'):
            return topic1 + topic2
        return topic1 + '/' + topic2

    @enveloped
    @parse_envelope('topic')
    @parse_envelope('retain')
    @parse_envelope('qos')
    def _push(self, topic, retain, qos, envelope, payload):  # pylint: disable=arguments-differ
        if topic is None:
            raise ValueError("Topic was not defined either by the __init__ nor by the envelope")

        if not self.multi:
            self._publish(payload, topic, retain, qos)
        else:
            if not isinstance(payload, dict):
                raise TypeError("In multi mode the payload is required to be a dictionary")
            for k, v in payload.items():
                key_topic = self._topic_concat(topic, k)
                try:
                    self._publish(v, key_topic, retain, qos)
                except:  # pylint: disable=bare-except
                    self.logger.exception(
                        "Publishing failed for message on '%s' @ "
                        "%s:%s with qos=%s. Payload='%s'",
                        key_topic, self.host, self.port, qos, v
                    )

        return {'data': payload, **envelope} if envelope else payload


MQTTPush = Publish



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pnp',
 'pnp.api',
 'pnp.api.endpoints',
 'pnp.config',
 'pnp.console',
 'pnp.engines',
 'pnp.plugins',
 'pnp.plugins.pull',
 'pnp.plugins.push',
 'pnp.plugins.udf',
 'pnp.shared']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=2.11.2,<3.0.0',
 'aiohttp>=3.6.2,<4.0.0',
 'async_generator>=1.10,<2.0',
 'asyncio>=3.4.3,<4.0.0',
 'asyncws>=0.1,<0.2',
 'binaryornot>=0.4.4,<0.5.0',
 'cachetools>=4.1.0,<5.0.0',
 'click>=7.1.2,<8.0.0',
 'coloredlogs>=14.0,<15.0',
 'cronex>=0.1.3,<0.2.0',
 'dictmentor>=0.2.2,<0.3.0',
 'fastapi>=0.61.2,<0.62.0',
 'fastcore>=1.3.13,<2.0.0',
 'glom>=19.10.0,<20.0.0',
 'influxdb>=5.3.0,<6.0.0',
 'paho-mqtt>=1.5.0,<2.0.0',
 'pathspec>=0.8.0,<0.9.0',
 'psutil>=5.7.0,<6.0.0',
 'pytest-mock>=3.3.1,<4.0.0',
 'python-box<=3.4.6',
 'pytz>=2020.1,<2021.0',
 'pyyaml-include>=1.2.post2,<2.0',
 'requests>=2.23.0,<3.0.0',
 'schedule>=0.6.0,<0.7.0',
 'schema>=0.7.2,<0.8.0',
 'slacker>=0.14.0,<0.15.0',
 'starlette_exporter>=0.6.0,<0.7.0',
 'sty>=1.0.0-beta.12,<2.0.0',
 'syncasync>=20180812,<20180813',
 'typeguard>=2.7.1,<3.0.0',
 'typing-extensions>=3.7.4,<4.0.0',
 'tzlocal>=2.0.0,<3.0.0',
 'uvicorn>=0.12.2,<0.13.0']

extras_require = \
{'dht': ['Adafruit_DHT>=1.3.2,<2.0.0'],
 'dropbox': ['dropbox>=9.0.0,<10.0.0', 'urllib3>=1.20,<2.0'],
 'faceR': ['face-recognition>=1.2.2,<2.0.0', 'image>=1.5.24,<2.0.0'],
 'fitbit': ['fitbit>=0.3.0,<0.4.0'],
 'fritz:python_version >= "3.6" and python_version < "4.0"': ['fritzconnection>=1.2.0,<2.0.0'],
 'fswatcher': ['watchdog>=0.8.3,<0.9.0'],
 'ftp': ['pyftpdlib>=1.5.0,<2.0.0'],
 'gpio': ['RPi.GPIO>=0.6.5,<0.7.0'],
 'miflora': ['miflora>=0.4.0,<0.5.0'],
 'sound': ['numpy>=1.16.0,<2.0.0',
           'PyAudio>=0.2.11,<0.3.0',
           'scipy>=1.2.0,<2.0.0'],
 'speedtest': ['speedtest-cli>=2.1.3,<3.0.0']}

entry_points = \
{'console_scripts': ['pnp = pnp.console.pnp:main',
                     'pnp_gmail_tokens = pnp.console.pnp_gmail_tokens:main',
                     'pnp_record_sound = pnp.console.pnp_record_sound:main']}

setup_kwargs = {
    'name': 'pnp',
    'version': '0.28.0',
    'description': "Pull 'n' Push",
    'long_description': '# Pull \'n\' Push\n\n[![Python](https://img.shields.io/badge/Python-3.6%20%7C%203.7%20%7C%203.8-green.svg)](https://www.python.org/)\n[![PyPI version](https://badge.fury.io/py/pnp.svg)](https://badge.fury.io/py/pnp)\n[![Docs](https://readthedocs.org/projects/pnp/badge/?version=stable)](https://pnp.readthedocs.io/en/stable/?badge=stable)\n[![GitHub Activity](https://img.shields.io/github/commit-activity/y/HazardDede/pnp.svg)](https://github.com/HazardDede/pnp/commits/master)\n[![Build Status](https://travis-ci.org/HazardDede/pnp.svg?branch=master)](https://travis-ci.org/HazardDede/pnp)\n[![Coverage Status](https://coveralls.io/repos/github/HazardDede/pnp/badge.svg?branch=master)](https://coveralls.io/github/HazardDede/pnp?branch=master)\n[![Docker: hub](https://img.shields.io/badge/docker-hub-brightgreen.svg)](https://cloud.docker.com/u/hazard/repository/docker/hazard/pnp)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n![Project Maintenance](https://img.shields.io/badge/maintainer-Dennis%20Muth%20%40HazardDede-blue.svg)\n\n> Pulls data from sources and pushes it to sinks with optional transformations in between.\n\n## Installation\n\n    pip install pnp\n\nInstallation with extras:\n    \n    pip install pnp[fswatcher,faceR]\n\nPlease consult the [component documentation](https://pnp.readthedocs.io/en/stable/plugins/index.html) to see if a\ncomponent requires an extra or not.\n\n## Getting started\n\nDefine `pulls` to fetch / pull data from source systems.\nDefine one `push` or multiple `pushes` per pull to transfer the pulled data anywhere else (you only need a plugin that \nknows how to handle the target). You configure your pipeline in `yaml`:\n\n```yaml\ntasks:\n  - name: hello-world\n    pull:\n      plugin: pnp.plugins.pull.simple.Repeat\n      args:\n        interval: 1s\n        repeat: "Hello World"\n    push:\n      - plugin: pnp.plugins.push.simple.Echo\n```\n        \nCopy this configuration and create the file `helloworld.yaml`. Run it:\n\n    pnp helloworld.yaml\n\nThis example yields the string `Hello World` every second.\n\n**Hint**: You can validate your config without actually executing it with\n\n```bash\n   pnp --check helloworld.yaml\n```\n\nIf you want to learn more please see the documentation at [Read the Docs](https://pnp.readthedocs.io/en/stable/).\n',
    'author': 'Dennis Muth',
    'author_email': 'd.muth@gmx.net',
    'maintainer': 'Dennis Muth',
    'maintainer_email': 'd.muth@gmx.net',
    'url': 'https://pnp.readthedocs.io/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""Utility / helper functions, classes, decorators, ..."""

# pylint: disable=too-many-lines
import inspect
import logging
import os
import re
import sys
import time
from base64 import b64encode
from collections import OrderedDict
from datetime import datetime, timedelta
from functools import wraps
from threading import Timer
from typing import (Union, Any, Optional, Iterable, Pattern, Dict, Callable, cast, Set, List)

from binaryornot.check import is_binary  # type: ignore
from box import Box, BoxKeyError  # type: ignore
from fastcore.basics import basic_repr
from typeguard import typechecked

from pnp import validator
from pnp.typing import DurationLiteral

FILE_MODES = ['binary', 'text', 'auto']

HTTP_METHODS = ['GET', 'HEAD', 'POST', 'PUT', 'DELETE', 'CONNECT', 'OPTIONS', 'TRACE', 'PATCH']


_LOGGER = logging.getLogger(__name__)


PY37 = sys.version_info >= (3, 7)


class EvaluationError(Exception):
    """Is raised when a selector expression cannot be evaluated."""


class StopCycleError(Exception):
    """A callback of interruptable_sleep should call this, when the sleep should be interrupted."""


def make_list(item_or_items: Any) -> Optional[List[Any]]:
    """
    Makes a list out of the given items.
    Examples:
        >>> make_list(1)
        [1]
        >>> make_list('str')
        ['str']
        >>> make_list(('i', 'am', 'a', 'tuple'))
        ['i', 'am', 'a', 'tuple']
        >>> print(make_list(None))
        None
        >>> # An instance of lists is unchanged
        >>> l = ['i', 'am', 'a', 'list']
        >>> l_res = make_list(l)
        >>> l_res
        ['i', 'am', 'a', 'list']
        >>> l_res is l
        True

    Args:
        item_or_items: A single value or an iterable.
    Returns:
        Returns the given argument as an list.
    """
    if item_or_items is None:
        return None
    if isinstance(item_or_items, list):
        return item_or_items
    if isinstance(item_or_items, dict):
        return [item_or_items]
    if hasattr(item_or_items, '__iter__') and not isinstance(item_or_items, str):
        return list(item_or_items)
    return [item_or_items]


def is_hashable(candidate: Any) -> bool:
    """
    Determines whether candidate can be hashed or not.

    Args:
        candidate: The candidate to test if it is hashable.

    Returns:
        True if `candidate` is hashable; otherwise False.

    Examples:

        >>> is_hashable("i am")
        True
        >>> is_hashable({"I am": "not"})
        False
    """
    try:
        hash(candidate)
    except TypeError:
        return False
    return True


def make_hashable(obj: Any) -> Any:
    """
    Converts a non-hashable instance into a hashable instance. Will take care of nested
    objects (like in iterables, dictionaries) as well. Will not detect a recursion
    and the function will fail in that case.

    Args:
        obj: The object to convert to a hashable object.

    Returns:
        Returns the hashable representation of the passed argument.

    Examples:

        >>> make_hashable("unchanged")
        'unchanged'
        >>> make_hashable((1, 2, 3))
        frozenset({1, 2, 3})
        >>> make_hashable({1: {2: [3, 4, 5]}})
        frozenset({(1, frozenset({(2, frozenset({3, 4, 5}))}))})
    """
    if isinstance(obj, dict):
        return frozenset({
            make_hashable(k): make_hashable(v)
            for k, v in obj.items()
        }.items())
    if is_iterable_but_no_str(obj):
        return frozenset([make_hashable(item) for item in obj])

    return obj if is_hashable(obj) else str(obj)


@typechecked
def camel_to_snake(name: str) -> str:
    """
    Converts camelCase to snake_case.
    https://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-snake-case

    >>> camel_to_snake('CamelCase')
    'camel_case'
    >>> camel_to_snake('CamelCamelCase')
    'camel_camel_case'
    >>> camel_to_snake('Camel2Camel2Case')
    'camel2_camel2_case'
    >>> camel_to_snake('getHTTPResponseCode')
    'get_http_response_code'
    >>> camel_to_snake('get2HTTPResponseCode')
    'get2_http_response_code'
    >>> camel_to_snake('HTTPResponseCode')
    'http_response_code'
    >>> camel_to_snake('HTTPResponseCodeXYZ')
    'http_response_code_xyz'

    """
    _str = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', _str).lower()


@typechecked
def wildcards_to_regex(
    wildcard_patterns: Union[str, Iterable[str]]
) -> Union[Pattern[str], Iterable[Pattern[str]]]:
    """
    Examples:

        >>> import fnmatch, re
        >>> (wildcards_to_regex(['sensor.*', 'sensor.lamp']) == [
        ...     re.compile(fnmatch.translate('sensor.*')),
        ...     re.compile(fnmatch.translate('sensor.lamp'))
        ... ])
        True
        >>> wildcards_to_regex('sensor.lamp') == re.compile(fnmatch.translate('sensor.lamp'))
        True
        >>> print(wildcards_to_regex(None))  #doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        TypeError: ... "wildcard_patterns" must be one of (str, Iterable); got NoneType instead
        >>> wildcards_to_regex([])
        []
    """

    import fnmatch
    if is_iterable_but_no_str(wildcard_patterns):
        return [re.compile(fnmatch.translate(str(item))) for item in wildcard_patterns]
    wildcard_patterns = cast(str, wildcard_patterns)
    return re.compile(fnmatch.translate(wildcard_patterns))


@typechecked
def include_or_exclude(
        item: str, include_regex: Optional[Iterable[Pattern[str]]] = None,
        exclude_regex: Optional[Iterable[Pattern[str]]] = None
) -> bool:
    """

    Returns:
        True if the item should be included; False if the item should be excluded.

    Examples:
        >>> re_all = wildcards_to_regex('sensor.*')
        >>> re1 = wildcards_to_regex('sensor.light1')
        >>> re2 = wildcards_to_regex('sensor.light2')

        >>> include_or_exclude('sensor.light1', [re1, re2])
        True
        >>> include_or_exclude('sensor.light1', [re1, re2], [re_all])
        False
        >>> include_or_exclude('sensor.light1', None, [re1])
        False
        >>> include_or_exclude('sensor.light2', None, [re1])
        True
        >>> include_or_exclude('sensor.light2', [re1], None)
        False

    """
    for regex in exclude_regex or []:
        if regex.match(item):
            return False
    for regex in include_regex or []:
        if regex.match(item):
            return True

    return include_regex is None


def transform_dict_items(dct, keys_fun=None, vals_fun=None):  # type: ignore
    """
    Transforms keys and/or values of the given dictionary by applying the given function.

    >>> dct = {"CamelCase": "gnaaa", "foo_oool": 42}
    >>> (transform_dict_items(dct, keys_fun=camel_to_snake) ==
    ...     {"camel_case": "gnaaa", "foo_oool": 42})
    True
    >>> transform_dict_items(dct, vals_fun=str) == {"CamelCase": "gnaaa", "foo_oool": "42"}
    True
    >>> (transform_dict_items(dct, keys_fun=camel_to_snake, vals_fun=str) ==
    ...     {"camel_case": "gnaaa", "foo_oool": "42"})
    True
    """
    if not keys_fun and not vals_fun:
        return dct

    def apply(kov: Any, fun: Optional[Callable[[Any], Any]]) -> Any:
        return kov if not fun else fun(kov)

    return {apply(k, keys_fun): apply(v, vals_fun) for k, v in dct.items()}


def is_iterable_but_no_str(candidate: Any) -> bool:
    """
    Checks if the given candidate is an iterable but not a str instance

    Example:
        >>> is_iterable_but_no_str(['a'])
        True
        >>> is_iterable_but_no_str('a')
        False
        >>> is_iterable_but_no_str(None)
        False
    """
    return hasattr(candidate, '__iter__') and not isinstance(candidate, (str, bytes))


def interruptable_sleep(wait: float, callback: Callable[[], None], interval: float = 0.5) -> None:
    """
    Waits the specified amount of time. The waiting can be interrupted when the callback raises a
    `StopCycleError`. The argument `interval` defines after how much wait time the callback
    should be called.

    Examples:

        # Does perform two cycles of 0.2 secs and one 0.1
        >>> interruptable_sleep(0.5, lambda: print("Cycle"), 0.2)
        Cycle
        Cycle
        >>> i = 0
        >>> def callback():
        ...     global i
        ...     if i == 2:
        ...         raise StopCycleError()
        ...     print("Cycle")
        ...     i = i + 1
        >>> interruptable_sleep(0.5, callback, 0.1)  # Aborts the wait after 2 cycles
        Cycle
        Cycle

    Args:
        wait (float): Wait for this amount of time, if not interrupted (same semantics as in
            `time.sleep(wait)`.
        callback: Callback that should raise an `StopCycleError` if the sleeps should be
            interrupted.
        interval (float): How often the callback should be called.

    Returns:
        None
    """
    def _cycles() -> int:
        return int(float(wait) // float(interval))

    complete_cycles = _cycles()
    try:
        for _ in range(0, complete_cycles):
            callback()  # Should raise a StopCycleError error when waiting should be aborted
            time.sleep(interval)

        time.sleep(wait % interval)
    except StopCycleError:
        pass


def sleep_until_interrupt(sleep_time: float, interrupt_fun: Callable[[], bool],
                          interval: float = 0.5) -> None:
    """Call this method to sleep an interruptable sleep until the interrupt function returns
    True."""
    validator.is_function(interrupt_fun=interrupt_fun)

    def callback() -> None:
        if interrupt_fun():
            raise StopCycleError()
    interruptable_sleep(sleep_time, callback, interval=interval)


def make_public_protected_private_attr_lookup(attr_name: str, as_dict: bool = False) \
        -> Union[Dict[str, str], List[str]]:
    """
    Given an attribute name this function will generate names of public, private and protected
    attribute names. The order is of lookups is always the given attr_name first and then
    descending by visibility (public -> proctected -> private)

    Examples:
        >>> make_public_protected_private_attr_lookup('my_lookup')  # Public attribute name
        ['my_lookup', '_my_lookup', '__my_lookup']
        >>> make_public_protected_private_attr_lookup('_my_lookup')  # Protected attribute name
        ['_my_lookup', 'my_lookup', '__my_lookup']
        >>> make_public_protected_private_attr_lookup('__my_lookup')  # Private attribute name
        ['__my_lookup', 'my_lookup', '_my_lookup']
        >>> l = make_public_protected_private_attr_lookup('_my_lookup', as_dict=True)
        >>> list(l.keys()), list(l.values())
        (['protected', 'public', 'private'], ['_my_lookup', 'my_lookup', '__my_lookup'])

    """
    validator.is_instance(str, lookup_name=attr_name)
    as_dict = try_parse_bool(as_dict, default=False)
    if attr_name.startswith('__'):
        # __lookup, lookup, _lookup
        res = OrderedDict([
            ('private', attr_name), ('public', attr_name[2:]), ('protected', attr_name[1:])
        ])
    elif attr_name.startswith('_'):
        # _lookup, lookup, __lookup
        res = OrderedDict([
            ('protected', attr_name), ('public', attr_name[1:]), ('private', '_' + attr_name)
        ])
    else:
        # lookup, _lookup, __lookup
        res = OrderedDict([
            ('public', attr_name), ('protected', '_' + attr_name), ('private', '__' + attr_name)
        ])
    return res if as_dict else list(res.values())


def is_local(server: str, allow_ipv6: bool = False) -> bool:
    """
    Checks if the given server (name or ip address) is actually a local.

    Examples:

        >>> is_local('www.google.de')
        False
        >>> is_local('LOCALHOST')
        True
        >>> is_local('127.0.0.1')
        True
        >>> is_local('0.0.0.0')
        True
        >>> is_local('::1')
        False
        >>> is_local('::1', allow_ipv6=True)
        True
    """
    server = str(server)
    return (
        server.lower() == 'localhost'
        or server == '127.0.0.1'
        or server == '0.0.0.0'
        or (server == '::1' and allow_ipv6)  # IPv6
    )


def safe_eval(source: str, **context: Any) -> Any:
    """
    Calls the `eval` function with modified globals() and locals(). By default of `eval` the
    current globals() and locals() are passed. We do not want that behaviour because the user
    can make a lot of nasty things with it (think of a already imported `os` module...).

    Instead we pass some functions (not all) from `__builtins__` (aka the whitelist) and remove
    all other modules. The caller can pass some additional context (like variables) to the eval.

    Examples:

        >>> payload = {'foo': {'bar': 1, 'baz': 2}}
        >>> safe_eval("str(payload['foo']['bar'])", payload=payload, str=str)
        '1'
        >>> import os
        >>> safe_eval("os.system('echo 0')")  # Try to access os
        Traceback (most recent call last):
        ...
        pnp.utils.EvaluationError: Failed to evaluate 'os.system('echo 0')'
        >>> safe_eval("os.system('echo 0')", os=os)
        0

    Args:
        source: The source to execute.
        **context: Additional context (e.g. variables) to pass to eval.

    Returns:
        Return the result of the eval. If something went wrong it raises an error.
    """
    try:
        return eval(str(source), {'__builtins__': {}}, context)  # pylint: disable=eval-used
    except Exception as exc:  # pylint: disable=bare-except
        raise EvaluationError("Failed to evaluate '{source}'".format(**locals())) from exc


def try_parse_int(candidate: Any) -> Optional[int]:
    """
    Convert the given candidate to int. If it fails None is returned.

    Example:

        >>> type(try_parse_int(1))  # int will still be int
        <class 'int'>
        >>> type(try_parse_int("15"))
        <class 'int'>
        >>> print(try_parse_int("a"))
        None

    Args:
        candidate: The candidate to convert.

    Returns:
        Returns the converted candidate if convertible; otherwise None.

    """
    try:
        return int(candidate)
    except (ValueError, TypeError):
        return None


def try_parse_int_float_str(candidate: Any) -> Union[float, int, str]:
    """
    Tries to parse the given value as an int or float. If this does not work simply the
    string-representation of the candidate will be returned.

    Examples:
        >>> res = try_parse_int_float_str(5)
        >>> res, type(res)
        (5, <class 'int'>)
        >>> res = try_parse_int_float_str(5.5)
        >>> res, type(res)
        (5.5, <class 'float'>)
        >>> res = try_parse_int_float_str(5.0)
        >>> res, type(res)
        (5, <class 'int'>)
        >>> res = try_parse_int_float_str('simple_string')
        >>> res, type(res)
        ('simple_string', <class 'str'>)

    Args:
        candidate: Candidate to convert.

    """
    try:
        ffloat = float(candidate)
        return int(ffloat) if ffloat.is_integer() else ffloat
    except:  # pylint: disable=bare-except
        return str(candidate)


def try_parse_bool(value: Any, default: Optional[Any] = None) -> Any:  # pylint: disable=too-many-return-statements
    """
    Tries to parse the given value as a boolean. If the parsing is unsuccessful the default will
    be returned.

    Args:
        value: Value to parse.
        default (optional): The value to return in case the conversion is not successful.

    Returns:
        If successful the converted representation of value; otherwise the default.

    Examples:
        >>> print(try_parse_bool(1))
        True
        >>> print(try_parse_bool('true'))
        True
        >>> print(try_parse_bool('T'))
        True
        >>> print(try_parse_bool('unknown', default='Unknown'))
        Unknown
        >>> print(try_parse_bool(None, default=True))
        True
        >>> print(try_parse_bool(1.0))
        True
        >>> print(try_parse_bool(0.99))
        True
        >>> print(try_parse_bool(0.0))
        False
        >>> print(try_parse_bool(lambda x: True, default="default"))
        True
    """
    if value is None:
        return default
    if isinstance(value, bool):
        return value
    if isinstance(value, str):
        if value.lower() in ['t', 'true', '1', 'wahr', 'ja', 'yes', 'on']:
            return True
        if value.lower() in ['f', 'false', '0', 'falsch', 'nein', 'no', 'off']:
            return False
        return default

    try:
        return bool(value)
    except (ValueError, TypeError):
        return default


def on_off(value: Any) -> str:
    """
    Tries to convert the given value to a on / off literal. If conversion is not possible
    (unsupported datatype) it will silently fail and return 'off'.

    Examples:

        >>> on_off(True), on_off(False)
        ('on', 'off')
        >>> on_off('True'), on_off('False')
        ('on', 'off')
        >>> on_off(None)
        'off'
        >>> on_off(99), on_off(0)
        ('on', 'off')

    Args:
        value: The value to convert to on / off.

    Returns:
        Return 'on' if the the `value` is True or equivalent (see try_parse_bool);
        otherwise 'off'.
    """
    return 'on' if try_parse_bool(value, default=False) else 'off'


def bps_mbps(val: float) -> float:
    """
    Converts bits per second (bps) into megabits per second (mbps).

    Examples:

        >>> bps_mbps(1000000)
        1.0
        >>> bps_mbps(1129000)
        1.13

    Args:
        val (float): The value in bits per second to convert.

    Returns:
        Returns val in megabits per second.
    """
    return round(float(val) / 1000000, 2)


def get_file_mode(file_path: str, mode: str) -> str:
    """
    Returns 'rb' if mode = 'binary'.
    Returns 'r' if mode = 'text'
    Returns 'rb' or 'r' if mode = 'auto' -> Will be automatically detected.

    Example:

        >>> get_file_mode("doesn't matter", 'binary')
        'rb'
        >>> get_file_mode("doesn't matter", 'text')
        'r'
        >>> get_file_mode(__file__, 'auto')
        'r'

    Args:
        file_path: File to load.
        mode: One of ['binary', 'text', 'auto'].

    Returns:
        One of ['rb', 'r']

    """
    if mode == 'binary':
        return 'rb'
    if mode == 'text':
        return 'r'
    if mode == 'auto':
        return 'rb' if is_binary(file_path) else 'r'

    raise ValueError("Argument 'mode' is expected to be one of: auto, binary, text")


def load_file(fp: str, mode: str = 'auto', base64: bool = False) -> Dict[str, Any]:
    """
    Loads a file by the given file path into memory.
    The read mode can either be binary, text or auto. Auto will try to guess the read mode
    based on the files content.

    Args:
        fp (str): file path.
        mode (str): One of binary, text, auto
        base64 (bool): If True the content will be base64 encoded; otherwise not.
    """
    if not isinstance(base64, bool):
        raise TypeError("Argument 'base64' is expected to be a bool, but is {}".format(
            type(base64)
        ))
    mode = 'binary' if base64 else mode
    if mode not in FILE_MODES:
        raise ValueError("Argument 'mode' is expected to be one of: {}".format(FILE_MODES))

    read_mode = get_file_mode(fp, mode=mode)
    with open(fp, read_mode) as fstream:
        contents = b64encode(fstream.read()) if base64 else fstream.read()

    return dict(file_name=os.path.basename(fp), content=contents, mode=read_mode, base64=base64)


def get_first_existing_file(*file_list: str) -> Optional[str]:
    """
    Given is a list of possible files. The function returns the first item that exists.
    If none of the candidates exists `None` is returned.

    Examples:
        >>> import tempfile
        >>> with tempfile.NamedTemporaryFile() as ntf:
        ...     res = get_first_existing_file('idonotexists.exists', ntf.name, '/asd/dasd/ad.ad')
        >>> res == ntf.name
        True
    """
    for fp in file_list:
        if isinstance(fp, str) and os.path.isfile(fp):
            return fp
    return None


def get_bytes(stream_or_file: Any) -> Union[bytes, str]:
    """
    Returns the bytes of the given stream or file argument. If it is a file, it will be opened
    in binary mode.

    Examples:
        >>> import tempfile
        >>> tmp = tempfile.NamedTemporaryFile()
        >>> print(get_bytes(tmp.name))
        b''
        >>> with open(tmp.name, 'rb') as stream:
        ...     get_bytes(stream)
        b''
        >>> get_bytes('thatonedoesnotexist')
        Traceback (most recent call last):
            ...
        ValueError: Argument 'stream_or_file' is neither a stream nor a file

    Args:
        stream_or_file: Stream or file to read the bytes from.

    Returns:
        Bytes of ``stream_or_file``
    """
    if getattr(stream_or_file, 'read', None):
        return cast(Union[bytes, str], stream_or_file.read())
    from pathlib import Path
    if Path(stream_or_file).is_file():
        with open(stream_or_file, 'rb') as file:
            return file.read()

    raise ValueError("Argument 'stream_or_file' is neither a stream nor a file")


def is_real_float(candidate: Any) -> bool:
    """
    Checks if the given candidate is a real float. An integer will return False.

    Examples:
        >>> is_real_float(1.1)
        True
        >>> is_real_float(1.0)
        False
        >>> is_real_float(object())
        False
        >>> is_real_float(1)
        False
        >>> is_real_float("str")
        False
    """
    try:
        return not float(candidate).is_integer()
    except (TypeError, ValueError):
        return False


def parse_duration_literal(literal: DurationLiteral) -> int:
    """
    Converts duration literals as '1m', '1h', and so on to an actual duration in seconds.
    Supported are 's' (seconds), 'm' (minutes), 'h' (hours), 'd' (days) and 'w' (weeks).

    Examples:

        >>> parse_duration_literal(60)  # Int will be interpreted as seconds
        60
        >>> parse_duration_literal(60.5)  # Float gets truncated
        60
        >>> parse_duration_literal('10')  # Any int convertible will be interpreted as seconds
        10
        >>> parse_duration_literal('20s')  # Seconds literal
        20
        >>> parse_duration_literal('2m')  # Minutes literal
        120
        >>> parse_duration_literal('1h')  # Hours literal
        3600
        >>> parse_duration_literal('1d')  # Days literal
        86400
        >>> parse_duration_literal('1w')  # Weeks literal
        604800
        >>> parse_duration_literal('invalid')  # Invalid will raise an error
        Traceback (most recent call last):
        ...
        TypeError: Interval 'invalid' is not a valid literal

    Args:
        literal: Literal to parse.

    Returns:
        Returns the converted literal's duration in seconds. If conversion is not possible
        an exception is raised.
    """
    try:
        # if successful we got seconds
        return int(literal)
    except:  # pylint: disable=bare-except
        # We have to check for s, m, h, d, w suffix
        seconds_per_unit = {"s": 1, "m": 60, "h": 3600, "d": 86400, "w": 604800}
        # Remove all non-alphanumeric letters
        _str = re.sub('[^0-9a-zA-Z]+', '', str(literal))
        value_str, unit = _str[:-1], _str[-1].lower()
        value = try_parse_int(value_str)
        if value is None or unit not in seconds_per_unit:
            raise TypeError("Interval '{}' is not a valid literal".format(literal)) from None
        return value * seconds_per_unit[unit]


def parse_duration_literal_float(literal: DurationLiteral) -> float:
    """
    Converts duration literals as '1m', '1h', and so on to an actual duration in seconds.
    Supported are 's' (seconds), 'm' (minutes), 'h' (hours), 'd' (days) and 'w' (weeks).

    Difference between `parse_duration_literal` and `parse_duration_literal_float` that
    the `*_float` will return a float and therefore allow fractions

    Examples:
        >>> parse_duration_literal_float(60.5)  # Float will be preserved
        60.5
        >>> parse_duration_literal_float('20s')  # Seconds literal
        20.0

    Args:
        literal: Literal to parse.

    Returns:
        Returns the converted literal's duration in seconds. If conversion is not possible
        an exception is raised.
    """
    if is_real_float(literal):
        return float(literal)

    return float(parse_duration_literal(literal))


def get_field_mro(cls: type, field_name: str) -> Set[str]:
    """
    Goes up the mro (method resolution order) of the given class and returns the union of a given
    field.

    Example:

        >>> class Root:
        ...     __myfield__ = 'root'

        >>> class A(Root):
        ...     __myfield__ = ['a', 'common']

        >>> class B(Root):
        ...     __myfield__ = ['b', 'common']

        >>> class Final(A, B):
        ...     __myfield__ = 'final'

        >>> get_field_mro(Final, '__myfield__') == {'root', 'a', 'b', 'common', 'final'}
        True
        >>> get_field_mro(A, '__myfield__') == {'root', 'a', 'common'}
        True
        >>> f = Final()
        >>> get_field_mro(f, '__myfield__') == {'root', 'a', 'b', 'common', 'final'}
        True

    Args:
        cls:
        field_name:

    Returns:

    """
    res = set()  # type: Set[str]
    if not hasattr(cls, '__mro__'):
        # cls might be an instance. mro is only available on classes
        if not hasattr(type(cls), '__mro__'):
            # No class, no instance ... Return empty set
            return res
        cls = type(cls)

    for clazz in inspect.getmro(cls):
        values_ = getattr(clazz, field_name, None)
        if values_ is not None:
            res = res.union(set(cast(Iterable[str], make_list(values_))))

    return res


class classproperty(property):  # pylint: disable=invalid-name
    """
    Decorator classproperty:
    Make class methods look like read-only class properties.
    Writing to that classproperty will not do what you expect ;-)
    Examples:
        >>> class Foo(object):
        ...     _instance = 5
        ...     @classproperty
        ...     def my_prop(cls):
        ...         return cls._instance
        >>> Foo.my_prop
        5
        >>> Foo._instance
        5
        >>> Foo._instance = 15
        >>> Foo.my_prop
        15
        >>> Foo.my_prop = 10
        >>> Foo._instance
        15
    """
    def __get__(self, cls, owner):  # type: ignore
        return classmethod(self.fget).__get__(None, owner)()


class Loggable:
    """
    Adds a logger property to the class to provide easy access to a configured logging instance to
    use.

    Example:
        >>> class NeedsLogger(Loggable):
        ...     def do(self, message):
        ...         self.logger.info(message)
        >>> dut = NeedsLogger()
        >>> dut.do('mymessage')
    """
    @classproperty
    def logger(cls: Any) -> logging.Logger:  # pylint: disable=no-self-argument
        """
        Configures and returns a logger instance for further use.
        Returns:
            (logging.Logger)
        """
        component = "{}.{}".format(cls.__module__, cls.__name__)  # pylint: disable=no-member
        return logging.getLogger(component)


class ReprMixin:
    """Adds a __repr__ and a __str__ method to this instance. You can control the repr fields
    via the __REPR_FIELDS__ class instance.

    Examples:

        >>> class A(ReprMixin):
        ...     __REPR_FIELDS__ = ['a']
        ...     def __init__(self):
        ...         self.a = 13

        >>> class B(A):
        ...     __REPR_FIELDS__ = 'b'
        ...     def __init__(self):
        ...         super().__init__()
        ...         self.b = 42

        >>> repr(A())
        'A(a=13)'
        >>> repr(B())
        'B(a=13, b=42)'
    """

    def __repr__(self) -> str:
        repr_fields = sorted(list(get_field_mro(self.__class__, '__REPR_FIELDS__')))
        res = basic_repr(repr_fields)(self)
        assert isinstance(res, str)
        return res

    def __str__(self) -> str:
        return repr(self)


class FallbackBox(Box):  # type: ignore
    """
    In some cases we want to handle integers and floats as string when accessing
    a boxed dictionary. For example we would like to mimic access paths like the zway api can:
        devices[19].instances[0].commandClasses[49].data[3].val.value
    In this case all int looking ones are actual string keys. But for plugins we want it
    as easy as possible. So we first take the key as it is to access the data, but when it raises
    a `KeyError` we cast the key to str and try it again to access the key as a fallback.

    Examples:

        >>> d = {'devices': {'1': {'instances': {'0': 'instance0', 1: 'instance1'}}}}
        >>> dut = FallbackBox(d)
        >>> dut.devices[1].instances[0]
        'instance0'
        >>> dut.devices[1].instances[1]
        'instance1'
    """
    def __getitem__(self, item, _ignore_default=False):  # type: ignore
        try:
            return super().__getitem__(item, _ignore_default=_ignore_default)
        except BoxKeyError as kerr:
            if isinstance(item, (int, float)):
                return super().__getitem__(str(item), _ignore_default=_ignore_default)
            raise kerr


class Debounce:
    """
    Defers a function execution until `wait` seconds have elapsed since the last time it was
    invoked. It is a way to collect multiple invokes and only really execute the last invocation.

    Examples:

        >>> results = []
        >>> def foo(counter):
        ...     results.append(counter)

        >>> dut = Debounce(foo, 0.2)  # Defers the execution for 200 milliseconds
        >>> for cnt in range(3): dut(cnt)
        >>> time.sleep(0.3)
        >>> results
        [2]

        >>> for cnt in range(4): dut(cnt)
        >>> time.sleep(0.3)
        >>> results
        [2, 3]

        >>> results.clear()
        >>> for cnt in range(3): dut(cnt)
        >>> dut.execute_now()
        >>> for cnt in range(4): dut(cnt)
        >>> time.sleep(0.3)
        >>> results
        [2, 3]



    """
    def __init__(self, fun: Callable[..., Any], wait: float = 0.5):
        validator.is_function(fun=fun)
        self.fun = fun
        self.wait = float(wait)
        self.timer = None  # type: Optional[Timer]

    def _safe_stop_timer(self) -> None:
        if self.timer is not None and not self.timer.finished.is_set():  # type: ignore
            self.timer.cancel()
        self.timer = None

    def execute_now(self) -> None:
        """Execute the debounced function call now. No matter what."""
        # Without checking self.timer.finished.is_set(), in rare situations the actual function
        # would get executed twice.
        if self.timer is not None and not self.timer.finished.is_set():  # type: ignore
            self._on_timer(*self.timer.args, **self.timer.kwargs)  # type: ignore

    def _setup_timer(self, *args: Any, **kwargs: Any) -> None:
        self._safe_stop_timer()
        self.timer = Timer(self.wait, self._on_timer, list(args), kwargs)
        self.timer.start()

    def _on_timer(self, *args: Any, **kwargs: Any) -> None:
        self._safe_stop_timer()
        self.fun(*args, **kwargs)

    def __call__(self, *args: Any, **kwargs: Any) -> None:
        self._setup_timer(*args, **kwargs)


class Cooldown(Debounce):
    """Executes the function once. Succeeding calls will be ignored until the cool down
    interval is expired. Will emit a callback when the cool down interval has expired.

    Example:
        >>> results = []
        >>> def foo(counter):
        ...     results.append(counter)
        >>> def cool_down():
        ...     results.append("cool_down")

        >>> dut = Cooldown(foo, 0.1, cool_down, True)
        >>> for cnt in range(5): dut(cnt)
        >>> time.sleep(0.15)
        >>> results
        [0, 'cool_down']

        >>> results = []
        >>> dut = Cooldown(foo, 0.10, cool_down, False)
        >>> dut(1)
        >>> time.sleep(0.05)
        >>> dut(2)
        >>> results
        [1]
        >>> time.sleep(0.03)
        >>> results
        [1]
        >>> time.sleep(0.03)
        >>> results
        [1, 'cool_down']
        >>> dut(3)
        >>> dut(4)
        >>> dut(5)
        >>> results
        [1, 'cool_down', 3]
        >>> time.sleep(0.11)
        >>> results
        [1, 'cool_down', 3, 'cool_down']
    """

    def __init__(
            self,
            fun: Callable[..., Any], cool_down: float = 0.5,
            cool_down_callback: Optional[Callable[[], None]] = None,
            renew_cooldown: bool = True
    ):
        super().__init__(fun, cool_down)
        if cool_down_callback:
            validator.is_function(cool_down_callback=cool_down_callback)
        self.cool_down_callback = cool_down_callback
        self.renew_cooldown = bool(renew_cooldown)

    def _setup_timer(self, *args: Any, **kwargs: Any) -> None:
        if self.renew_cooldown or self.timer is None:
            super()._setup_timer()

    def _on_timer(self, *args: Any, **kwargs: Any) -> None:
        self._safe_stop_timer()
        if self.cool_down_callback:
            self.cool_down_callback()

    def __call__(self, *args: Any, **kwargs: Any) -> None:
        if self.timer is None:  # No timer running -> first call
            self.fun(*args, **kwargs)
        self._setup_timer()  # Setup or renew the cooldown


class Singleton:
    """
    Derive from this class to create a singleton.
    Examples:
        >>> class Foo(Singleton):
        ...     def __init__(self, state):
        ...         self.state = state
        >>> # Instance will be created on first call ...
        >>> f = Foo('statef')
        >>> # ... and passing different arguments will not change the already created instance.
        >>> g = Foo('stateg')
        >>> f is g
        True
        >>> print(f.state)
        statef
        >>> print(g.state)
        statef
        >>> # Two ways to retrieve the singleton instance
        >>> # 1. Pretend to "create" a new instance. This will return the already created instance.
        >>> Foo() is f
        True
        >>> # Arguments work as well (but will be ignored)
        >>> Foo("i", "get", "ignored") is f
        True
        >>> # 2. The more explicit variant: Call ``instance`` on the Singleton (Explicit is good).
        >>> Foo.instance is f
        True
        >>> # You can re-initialize the instance
        >>> f.init('new_state')
        >>> print(f.state)
        new_state
        >>> print(g.state)
        new_state
    """
    def __new__(cls, *args: Any, **kwargs: Any) -> Any:
        instance = cls.__dict__.get("__singleton__")
        if instance is None:
            # Create the singleton
            cls.__singleton__ = instance = object.__new__(cls)
            # Initialize it once.
            instance.__init__(*args, **kwargs)

            # Backup initializer ...
            cls.init = cls.__init__
            # .. and then forget about it...

            def empty_init(self, *args, **kwargs):  # type: ignore # pylint: disable=unused-argument
                pass
            cls.__init__ = empty_init  # type: ignore
        return instance

    @classproperty
    def instance(cls: Any) -> Any:  # pylint: disable=no-self-argument
        """
        Returns the singleton instance. If there is none yet, it will be created by
        calling the `init` method without any arguments.
        """
        return cls.__new__(cls)


class FileLock:
    """
    A file locking mechanism that has context-manager support so you can use it within a with
    statement.
    This implementation should be relatively cross-platform compatible cause it doesn't rely on
    msvcrt or fcntl for the locking.
    You can nest calls to ``acquire`` it will only lock the file once and release it with the last
    call to ``release``.

    Examples:
        >>> import tempfile
        >>> tmp = tempfile.NamedTemporaryFile()
        >>> # Basic usage with context manager
        >>> with FileLock(tmp.name) as l:
        ...     # Do something nice with the file
        ...     print(l.locked)
        True
        >>> # Basic usage w/o context manager
        >>> l = FileLock(tmp.name)
        >>> l.locked
        False
        >>> l.acquire()
        >>> try:
        ...     # Do something nice with the file
        ...     print(l.locked)
        ... finally:
        ...     l.release()
        True
        >>> print(l.locked)
        False
        >>> # Lock counter
        >>> l = FileLock(tmp.name)
        >>> with l:
        ...     print(l.lock_counter)
        ...     with l:
        ...         print(l.lock_counter)
        ...     print((l.lock_counter, l.locked))
        1
        2
        (1, True)
        >>> print((l.lock_counter, l.locked))
        (0, False)
        >>> # Blocking
        >>> with FileLock(tmp.name) as l1:
        ...     with FileLock(tmp.name, timeout=1) as l2:
        ...         pass
        Traceback (most recent call last):
        ...
        pnp.utils.FileLock.TimeoutError: Timeout occurred
    """

    class TimeoutError(Exception):
        """Is raised when a timeout occurs when waiting for the file lock."""

    def __init__(self, file_name: str, timeout: float = 10, delay: float = .05):
        """
        Prepare the file locker. Specify the file to lock and optionally the maximum timeout
        (seconds) and the delay between each attempt to lock.

        Args:
            file_name: The file to lock exclusively.
            timeout: If the file is already locked by another process, this instance tries
                to acquire the lock for a maximum of `timeout` seconds.
            delay:
        """
        self.lock_counter = 0
        self.lockfile = os.path.join(os.getcwd(), "%s.lock" % file_name)
        self.file_name = file_name
        self.timeout = timeout
        self.delay = delay
        self.fd = None  # type: Optional[int]  # pylint: disable=invalid-name

    def acquire(self) -> None:
        """
        Acquires the lock, if possible. If the file is already locked, the algorithm tries to
        acquire the lock until it either gets the lock or the timeout threshold is exceeded.
        In case of a timeout an error is raised.

        Returns:
            None
        """
        if self.lock_counter > 0:
            # Dude, you already locked the file...
            self.lock_counter += 1
            return

        start_time = time.time()
        while True:
            try:
                self.fd = os.open(self.lockfile, os.O_CREAT | os.O_EXCL | os.O_RDWR)
                self.lock_counter += 1
                break
            except OSError as oserr:
                import errno
                if oserr.errno != errno.EEXIST:
                    raise
                if (time.time() - start_time) >= self.timeout:
                    raise FileLock.TimeoutError("Timeout occurred")
                time.sleep(self.delay)

    def release(self) -> None:
        """
        Unlocks the file when the exclusive lock is no longer needed.

        Returns:
            None
        """
        if self.lock_counter <= 0:
            # Dude, there is no lock...
            return

        self.lock_counter -= 1
        if self.lock_counter == 0:
            if self.fd:
                os.close(self.fd)
            os.unlink(self.lockfile)

    @property
    def locked(self) -> bool:
        """
        Checks whether the file is currently locked by any process.

        Returns: True if the file is currently locked (by any process, not just this one);
            otherwise False.
        """
        return os.path.isfile(self.lockfile)

    def __enter__(self) -> 'FileLock':
        """
        Context manager acquire lock.

        Returns:
            Self
        """
        self.acquire()
        return self

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        """
        Context manager release.

        Args:
            exc_type: unused
            exc_value: unused
            traceback: unused

        Returns:
            None
        """
        self.release()

    def __del__(self) -> None:
        """
        Make sure that this instance doesn't leave a lockfile lying around when garbage collected.

        Returns:
            None
        """
        self.release()


class Throttle:
    """
    Decorator that prevents a function from being called more than once every
    time period.

    Example:
        >>> @Throttle(timedelta(seconds=1))
        ... def my_fun():
        ...     return "Called"
        >>> my_fun()
        'Called'
        >>> my_fun(), my_fun()
        (None, None)
        >>> time.sleep(1)
        >>> my_fun()
        'Called'

    """
    def __init__(self, delta: timedelta):
        self.throttle_period = delta
        validator.is_instance(timedelta, delta=delta)
        self.time_of_last_call = datetime.min

    def __call__(self, fun: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(fun)
        def wrapper(*args: Any, **kwargs: Any) -> Optional[Any]:
            now = datetime.now()
            time_since_last_call = now - self.time_of_last_call

            if time_since_last_call > self.throttle_period:
                self.time_of_last_call = now
                return fun(*args, **kwargs)
            return None

        return wrapper



"""Fitbit related plugins."""

import asyncio
import os
import pathlib
from functools import partial
from typing import Iterable, Any

import schema
from ruamel import yaml

from pnp import validator
from pnp.plugins import load_optional_module
from pnp.plugins.pull import AsyncPolling
from pnp.utils import camel_to_snake, transform_dict_items, make_list, FileLock


class _FitbitBase(AsyncPolling):

    __REPR_FIELDS__ = ['_config', '_system']

    EXTRA = 'fitbit'

    TOKEN_SCHEMA = schema.Schema({
        'client_id': schema.Use(str),
        'client_secret': schema.Use(str),
        'access_token': schema.Use(str),
        'refresh_token': schema.Use(str),
        'expires_at': schema.Use(float)
    })

    def __init__(self, config, system=None, **kwargs):
        super().__init__(**kwargs)
        self._system = system

        # Absolute config path? Relative -> relative to pnp configuration file
        self._config = config
        if not os.path.isabs(config):
            self._config = os.path.join(self.base_path, config)
        validator.is_file(config=self._config)

        import threading
        self._client_lock = threading.Lock()
        self._tokens = None  # Set by _load_tokens
        self._tokens_tstamp = None  # Set by _load_tokens and _save_tokens
        self._client = None  # Set by client property

    @property
    def client(self):
        """Return the configured fitbit client."""
        with self._client_lock:
            if self._load_tokens():  # Check if tokens got refreshed by another process
                fbit = load_optional_module('fitbit', self.EXTRA)
                self._client = fbit.Fitbit(**self._tokens, oauth2=True,
                                           refresh_cb=self._save_tokens, system=self._system)
            return self._client

    def _load_tokens(self):
        current_tstamp = pathlib.Path(self._config).stat().st_mtime
        # Compare memorized modified timestamp (might be None initially) and current
        if self._tokens_tstamp == current_tstamp:
            # Tokens did not change. Skip
            return False

        self.logger.debug("Loading tokens from %s: Requesting lock", self._config)
        with FileLock(self._config):
            self.logger.debug("Loading tokens from %s: Lock acquired", self._config)
            with open(self._config, 'r') as fp:
                _tokens = yaml.safe_load(fp)
                self._tokens = self.TOKEN_SCHEMA.validate(_tokens)
            self._tokens_tstamp = current_tstamp
        self.logger.debug("Loading tokens from %s: Lock released", self._config)
        return True

    def _save_tokens(self, tokens):
        new_config = {
            'client_id': self._tokens.get('client_id'),
            'client_secret': self._tokens.get('client_secret'),
            'access_token': tokens.get('access_token'),
            'refresh_token': tokens.get('refresh_token'),
            'expires_at': tokens.get('expires_at')
        }
        self.logger.debug("Saving tokens to %s: Requesting lock", self._config)
        self.TOKEN_SCHEMA.validate(new_config)
        with FileLock(self._config):
            self.logger.debug("Saving tokens to %s: Lock acquired", self._config)
            with open(self._config, 'w') as fp:
                yaml.dump(new_config, fp, default_flow_style=False)
            self._tokens_tstamp = pathlib.Path(self._config).stat().st_mtime
            self._tokens = new_config
            self.logger.debug("Saving tokens to %s: Lock released", self._config)

    async def _poll(self):
        raise NotImplementedError()  # pragma: no cover


class Current(_FitbitBase):
    """
    Requests various latest / current metrics (steps, calories, distance, ...)
    from the fitbit api.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/pull/fitbit.Current/index.md
    """
    __REPR_FIELDS__ = '_resources'

    def __init__(self, resources: Iterable[str], **kwargs: Any):
        super().__init__(**kwargs)
        self._resource_map = self._create_resource_map()
        self._resources = make_list(resources)
        for res in self._resources:
            validator.one_of(list(self._resource_map.keys()), resource=res)

    def _create_resource_map(self):
        return {
            # Activities
            'activities/calories': partial(self._conv, conv_fun=int),
            'activities/caloriesBMR': partial(self._conv, conv_fun=int),
            'activities/steps': partial(self._conv, conv_fun=int),
            'activities/distance': partial(self._conv, conv_fun=float),
            'activities/floors': partial(self._conv, conv_fun=int),
            'activities/elevation': partial(self._conv, conv_fun=int),
            'activities/minutesSedentary': partial(self._conv, conv_fun=int),
            'activities/minutesLightlyActive': partial(self._conv, conv_fun=int),
            'activities/minutesFairlyActive': partial(self._conv, conv_fun=int),
            'activities/minutesVeryActive': partial(self._conv, conv_fun=int),
            'activities/activityCalories': partial(self._conv, conv_fun=int),
            # Body
            'body/bmi': partial(self._conv, conv_fun=float),
            'body/fat': partial(self._conv, conv_fun=float),
            'body/weight': partial(self._conv, conv_fun=float),
            # Foods
            'foods/log/caloriesIn': partial(self._conv, conv_fun=int),
            'foods/log/water': partial(self._conv, conv_fun=float),
            # Sleep
            'sleep/awakeningsCount': partial(self._conv, conv_fun=int),
            'sleep/efficiency': partial(self._conv, conv_fun=int),
            'sleep/minutesAfterWakeup': partial(self._conv, conv_fun=int),
            'sleep/minutesAsleep': partial(self._conv, conv_fun=int),
            'sleep/minutesAwake': partial(self._conv, conv_fun=int),
            'sleep/minutesToFallAsleep': partial(self._conv, conv_fun=int),
            'sleep/startTime': partial(self._conv, conv_fun=None),
            'sleep/timeInBed': partial(self._conv, conv_fun=int)
        }

    def _conv(self, resource, conv_fun):
        resp = self._api_fetch_current(resource)
        return conv_fun(resp) if conv_fun else resp

    def _api_fetch_current(self, resource):
        resp = self.client.time_series(resource, period='7d')
        first_key = next(iter(resp.values()))
        if first_key is None:
            return None  # pragma: no cover
        return first_key[-1].get('value')

    def _call(self, resource):
        return self._resource_map.get(resource)(resource)

    async def _poll(self):
        loop = asyncio.get_event_loop()
        coros = [loop.run_in_executor(None, self._call, res) for res in self._resources]
        results = await asyncio.gather(*coros)
        return transform_dict_items({
            res: results[i] for i, res in enumerate(self._resources)
        }, keys_fun=camel_to_snake)


class Devices(_FitbitBase):
    """
    Requests details (battery, model, ...) about your fitbit devices / trackers
    associated to your account.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/pull/fitbit.Devices/index.md
    """
    async def _poll(self):
        devices = await asyncio.get_event_loop().run_in_executor(None, self.client.get_devices)
        return [transform_dict_items(d, keys_fun=camel_to_snake) for d in devices]


class Goal(_FitbitBase):
    """
    Requests your goals (water, steps, ...) from the fitbit api.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/pull/fitbit.Goal/index.md
    """
    __REPR_FIELDS__ = '_goals'

    def __init__(self, goals, **kwargs):
        super().__init__(**kwargs)
        self._goals_map = self._create_goal_map()
        self._goals = make_list(goals)
        for goal in self._goals:
            validator.one_of(list(self._goals_map.keys()), goal=goal)

    def _create_goal_map(self):
        return {
            'body/fat': lambda goal: float(self.client.body_fat_goal().get('goal', {}).get('fat')),
            'body/weight': lambda goal: (
                float(self.client.body_weight_goal().get('goal', {}).get('weight'))
            ),
            'activities/daily/activeMinutes': self._get_daily_activity_goal,
            'activities/daily/caloriesOut': self._get_daily_activity_goal,
            'activities/daily/distance': self._get_daily_activity_goal,
            'activities/daily/floors': self._get_daily_activity_goal,
            'activities/daily/steps': self._get_daily_activity_goal,
            'activities/weekly/distance': self._get_weekly_activity_goal,
            'activities/weekly/floors': self._get_weekly_activity_goal,
            'activities/weekly/steps': self._get_weekly_activity_goal,
            'foods/calories': lambda goal: (
                int(self.client.food_goal().get('goals', {}).get('calories'))
            ),
            'foods/water': lambda goal: int(self.client.water_goal().get('goal', {}).get('goal'))
        }

    def _get_daily_activity_goal(self, goal):
        single_goal = goal.split('/')[-1]
        conv = float if single_goal == 'distance' else int
        return conv(self.client.activities_daily_goal().get('goals', {}).get(single_goal))

    def _get_weekly_activity_goal(self, goal):
        single_goal = goal.split('/')[-1]
        conv = float if single_goal == 'distance' else int
        return conv(self.client.activities_weekly_goal().get('goals', {}).get(single_goal))

    def _call(self, goal):
        return self._goals_map.get(goal)(goal)

    async def _poll(self):
        loop = asyncio.get_event_loop()
        coros = [loop.run_in_executor(None, self._call, goal) for goal in self._goals]
        results = await asyncio.gather(*coros)
        return transform_dict_items({
            goal: results[i] for i, goal in enumerate(self._goals)
        }, keys_fun=camel_to_snake)



"""Home assistant related user-defined functions."""

from pnp.plugins.udf import UserDefinedFunction
from pnp.shared.hass import HassApi


class State(UserDefinedFunction):
    """
    Fetches the state of an entity from home assistant by a rest-api request.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/udf/hass.State/index.md
    """
    __REPR_FIELDS__ = ['timeout', 'url']

    def __init__(self, url, token, timeout=10, **kwargs):
        super().__init__(**kwargs)
        self.url = str(url)
        self.token = str(token)
        self.timeout = timeout and int(timeout)
        if self.timeout <= 0:
            self.timeout = None  # Basically means no timeout
        self._client = HassApi(self.url, self.token, self.timeout)

    def action(self, entity_id, attribute=None):  # pylint: disable=arguments-differ
        """
        Calls the home assistant api to fetch the state of the given entity (or one of it's
        attributes).

        Args:
            entity_id (str): The entity id of the entity state of interest.
            attribute: Optionally specify an attribute of the entity.

        Returns:
            Returns the state of the entity itself or - if argument attribute is set - the state
            of the specified attribute of that entity.
        """
        entity_id = str(entity_id)
        attribute = attribute and str(attribute)
        endpoint = 'states/{entity_id}'.format(**locals())

        try:
            response = self._client.call(endpoint)
        except RuntimeError as exc:
            raise RuntimeError(
                "Failed to fetch the state for {entity_id} @ {self.url}".format(**locals())
            ) from exc

        if attribute:
            return response.get('attributes', {}).get(str(attribute))

        return response.get('state', None)



"""Custom logging handler for pnp."""

import asyncio
import traceback
from functools import partial
from logging import (
    Formatter,
    Handler,
    LogRecord,
    CRITICAL,
    DEBUG,
    ERROR,
    FATAL,
    INFO,
    NOTSET,
    WARNING,
    _nameToLevel
)
from typing import Optional, Iterable, Dict, Any, List

import slacker  # type: ignore
from typeguard import typechecked

from pnp import validator


class NoStacktraceFormatter(Formatter):
    """
    By default the stacktrace will be formatted as part of the message.
    Since we want the stacktrace to be in the attachment of the Slack message,
      we need a custom formatter to leave it out of the message.
    """

    def formatException(self, ei: Any) -> Any:
        return None


class SlackHandler(Handler):
    """
    Implements a logging handler that sends formatted messages to slack.
    """

    ERROR_COLOR = 'danger'  # color name is built in to Slack API
    WARNING_COLOR = 'warning'  # color name is built in to Slack API
    INFO_COLOR = '#439FE0'

    COLORS = {
        CRITICAL: ERROR_COLOR,
        FATAL: ERROR_COLOR,
        ERROR: ERROR_COLOR,
        WARNING: WARNING_COLOR,
        INFO: INFO_COLOR,
        DEBUG: INFO_COLOR,
        NOTSET: INFO_COLOR,
    }

    DEFAULT_EMOJI = ':robot:'
    DEFAULT_USERNAME = 'PnP'
    DEFAULT_PING_LEVEL = 'ERROR'
    MAX_ATTACHMENT_CHARS = 1000

    @typechecked
    def __init__(self, api_key: str, channel: str, username: str = DEFAULT_USERNAME,
                 icon_url: Optional[str] = None, icon_emoji: Optional[str] = None,
                 ping_users: Optional[Iterable[str]] = None,
                 ping_level: str = DEFAULT_PING_LEVEL, fire_and_forget: bool = True):
        super().__init__()
        self.formatter = NoStacktraceFormatter()
        self.slacker = slacker.Slacker(api_key)
        self.fire_and_forget = bool(fire_and_forget)
        self.username = str(username)
        self.icon_url = icon_url and str(icon_url)
        self.icon_emoji = icon_emoji if (icon_emoji or icon_url) else self.DEFAULT_EMOJI
        self.channel = channel
        if not self.channel.startswith('#') and not self.channel.startswith('@'):
            self.channel = '#' + self.channel

        # Optional ping user related stuff
        ping_level = str(ping_level).upper()
        validator.one_of(_nameToLevel, ping_level=ping_level)
        self.ping_level = _nameToLevel[ping_level]  # type: int
        self.ping_user_ids = []  # type: List[int]

        if ping_users:
            user_map = self._build_user_map()
            for ping_user in ping_users:
                user_id = user_map.get(ping_user)
                if not user_id:
                    raise RuntimeError(
                        'User not found in Slack users list: %s' % ping_user)
                self.ping_user_ids.append(user_id)

    def _build_user_map(self) -> Dict[str, int]:
        def _helper() -> Any:
            for usr in user_list:
                yield {usr['name']: usr['id']}
                yield {usr['profile']['real_name']: usr['id']}
                yield {usr['profile']['display_name']: usr['id']}

        user_list = self.slacker.users.list().body['members']
        return {k: v for d in _helper() for k, v in d.items()}

    def _build_message(self, record: LogRecord) -> str:
        return str(self.format(record))

    def _build_attachment(self, record: LogRecord) -> Dict[str, Any]:
        message = self._build_message(record)
        field = {
            'title': message,
            'short': False
        }
        if record.exc_info:
            trace = '\n'.join(traceback.format_exception(*record.exc_info))
            # Attachments have a character limit. They get truncated by slack
            # automagically, but this will also truncate the code blocks (```).
            # Se we do it on our own and we take the truncate the beginning cause
            # we assume that the end of the trace will contain the mor important
            # information.
            trace_size = len(trace) + 6  # 6 extra chars for backticks
            trace = (trace if trace_size <= self.MAX_ATTACHMENT_CHARS
                     else trace[trace_size - self.MAX_ATTACHMENT_CHARS:])
            field['value'] = "```{trace}```".format(trace=trace)
        return {
            'color': self.COLORS.get(record.levelno, self.INFO_COLOR),
            'fields': [field]
        }

    def emit(self, record: LogRecord) -> None:
        text = ''
        if self.ping_user_ids and record.levelno >= self.ping_level:
            for user in self.ping_user_ids:
                text = "<@{user}> {text}".format(user=user, text=text)

        attachments = [self._build_attachment(record)]

        # Use the executor to fire and forget about the message to post
        # and don't block the workflow just for logging...
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()

        slacker_call = partial(
            self.slacker.chat.post_message,
            text=text if text != '' else None,
            channel=self.channel,
            username=self.username,
            icon_url=self.icon_url,
            icon_emoji=self.icon_emoji,
            attachments=attachments
        )
        if self.fire_and_forget:
            loop.run_in_executor(None, slacker_call)  # Don't wait and don't check errors
        else:
            slacker_call()  # Wait for the call to finish; this will block asyncio



"""Pull 'n' Push Sound recorder."""
import sys

import click

CHANNELS = 1
RATE = 44100
CHUNK = 512


def list_devices(ctx, param, value):
    """List all available microphone devices."""
    _ = param  # Fake usage
    if not value or ctx.resilient_parsing:
        return

    try:
        import pyaudio
    except ImportError:
        print("You have to install extra 'sound' in order to use this shell script")
        ctx.exit(99)
        return  # Never reached, but linters do not know about it

    pa = pyaudio.PyAudio()
    for i in range(pa.get_device_count()):
        dev = pa.get_device_info_by_index(i)
        input_chn = dev.get('maxInputChannels', 0)
        if input_chn > 0:
            name = dev.get('name')
            rate = dev.get('defaultSampleRate')
            print("Index {i}: {name} (Max Channels {input_chn}, Default @ {rate} Hz)".format(
                i=i, name=name, input_chn=input_chn, rate=int(rate)

            ))
    ctx.exit()


@click.command('pnp_record_sound')
@click.argument(
    'output_file',
    type=click.Path(exists=False, dir_okay=False, resolve_path=True),
    required=True
)
@click.option(
    '--seconds',
    type=click.INT,
    default=1, show_default=True,
    help="Number of seconds to record."
)
@click.option(
    '--index',
    type=click.INT,
    default=None,
    help="Optionally specify the index of the mic to record the sound."
)
@click.option(
    '--list',
    is_flag=True,
    callback=list_devices,
    expose_value=False,
    is_eager=True,
    help="Shows all available mic devices and their associated index and exit."
)
def main(output_file, seconds, index):
    """Pull 'n' Push Sound recorder."""
    try:
        import pyaudio
        import wave
    except ImportError:
        print("You have to install extra 'sound' in order to use this shell script")
        sys.exit(99)

    fmt = pyaudio.paInt16
    pa = pyaudio.PyAudio()
    if index is not None and index > pa.get_device_count() - 1:
        print("Device index is out of bounds")
        sys.exit(1)
    if index is not None:
        device = pa.get_device_info_by_index(index)
        if device.get('maxInputChannels', 0) == 0:
            print("Device @ index {index} is not an input device (microphone)".format(**locals()))
            sys.exit(2)
        if device.get('maxInputChannels', 0) < CHANNELS:
            print("Device @ index {index} does not support {channels} channels".format(
                index=index, channels=CHANNELS
            ))
            sys.exit(3)

    buffer = []
    input("Press any key to start the recording")
    stream = pa.open(
        format=fmt,
        channels=CHANNELS,
        rate=RATE,
        input=True,
        input_device_index=index,
        frames_per_buffer=CHUNK
    )

    try:
        for _ in range(0, int(RATE / CHUNK * seconds)):
            data = stream.read(CHUNK)
            buffer.append(data)
    finally:
        stream.stop_stream()
        stream.close()
        pa.terminate()

    out_file = wave.open(output_file, 'wb')
    out_file.setnchannels(CHANNELS)
    out_file.setsampwidth(pa.get_sample_size(fmt))
    out_file.setframerate(RATE)
    out_file.writeframes(b''.join(buffer))
    out_file.close()

    print("Wave file written to {out_wav} @ {rate} Hz".format(out_wav=output_file, rate=RATE))


if __name__ == '__main__':
    main()  # pylint: disable=no-value-for-parameter



"""Contains utility methods for validating."""
import os
from typing import Any, Iterable, List, Optional

from typeguard import typechecked


@typechecked
def all_items(item_type: type, **kwargs: Iterable[Any]) -> None:
    """
    Examples:
        >>> all_items(str, a=['a', 'b', 'c'])
        >>> all_items(int, a=[1, 2, 3])
        >>> all_items(int, b=[1, 's', 3])
        Traceback (most recent call last):
        ...
        TypeError: Item value at pos 1 of argument 'b' is expected to be a <class 'int'>,\
 but is <class 'str'>
    """
    is_iterable_but_no_str(**kwargs)
    for arg_name, arg_val in kwargs.items():
        for i, item in enumerate(arg_val):
            if not isinstance(item, item_type):
                raise TypeError(
                    "Item value at pos {} of argument '{}' is expected to be a {}, "
                    "but is {}".format(
                        i, arg_name, item_type, type(item)
                    )
                )


def is_directory(**kwargs: Any) -> None:
    """
    Examples:
        >>> is_directory(arg='/tmp')
        >>> is_directory(arg='/thisonedoesnotexists')
        Traceback (most recent call last):
        ...
        ValueError: Argument 'arg' is expected to be a directory, but is '/thisonedoesnotexists'
    """
    for arg_name, arg_value in kwargs.items():
        if not os.path.isdir(arg_value):
            raise ValueError(
                "Argument '{arg_name}' is expected to be a directory, "
                "but is '{arg_value}'".format(arg_name=arg_name, arg_value=arg_value)
            )


def is_file(**kwargs: Any) -> None:
    """
    Examples:
        >>> import tempfile
        >>> with tempfile.NamedTemporaryFile() as tmpf:
        ...     is_file(arg=tmpf.name)
        >>> is_file(arg='/doesnotexist.txt')
        Traceback (most recent call last):
        ...
        ValueError: Argument 'arg' is expected to be a file, but is '/doesnotexist.txt'
    """
    for arg_name, arg_value in kwargs.items():
        if not os.path.isfile(str(arg_value)):
            raise ValueError(
                "Argument '{arg_name}' is expected to be a file, "
                "but is '{arg_value}'".format(
                    arg_name=arg_name, arg_value=arg_value
                )
            )


def is_function(**kwargs: Any) -> None:
    """
    Examples:
        >>> def foo():
        ...     pass
        >>> is_function(foo=foo)
        >>> is_function(bar='bar')
        Traceback (most recent call last):
        ...
        TypeError: Argument 'bar' is expected to be a function/callable, but is '<class 'str'>'
        >>> is_function(baz=lambda: True)
    """
    for arg_name, arg_value in kwargs.items():
        if not callable(arg_value):
            arg_type = type(arg_value)
            raise TypeError(
                "Argument '{arg_name}' is expected to be a function/callable, "
                "but is '{arg_type}'".format(
                    arg_name=arg_name, arg_type=arg_type
                )
            )


@typechecked
def is_instance(*required_types: type, allow_none: bool = False, **kwargs: Any) -> None:
    """
    Examples:

        >>> is_instance(str, arg="i am a string") #  Should be ok
        >>> is_instance(str, allow_none=True, arg=None)  # Should be ok as well
        >>> is_instance(bool, arg="i am a string")  # Not ok
        Traceback (most recent call last):
        ...
        TypeError: Argument 'arg' is expected to be a (<class 'bool'>,), but is <class 'str'>
        >>> is_instance(bool, str, arg="i am a string")  # Should be ok, too
        >>> is_instance(str, bool, arg1="i am a string", arg2=True)  # Ok ...
        >>> is_instance(str, arg1="i am a string", arg2=True)  # Not ok
        Traceback (most recent call last):
        ...
        TypeError: Argument 'arg2' is expected to be a (<class 'str'>,), but is <class 'bool'>
    """
    for arg_name, arg_value in kwargs.items():
        if not (allow_none and arg_value is None) and not isinstance(arg_value, required_types):
            raise TypeError(
                "Argument '{arg_name}' is expected to be a {required_type}"
                ", but is {actual_type}".format(
                    arg_name=arg_name,
                    required_type=required_types,
                    actual_type=type(arg_value)
                ))


def is_iterable_but_no_str(**kwargs: Any) -> None:
    """
    Examples:
        >>> is_iterable_but_no_str(l=[])
        >>> is_iterable_but_no_str(t=("a", "b"))
        >>> is_iterable_but_no_str(s="abc")
        Traceback (most recent call last):
        ...
        TypeError: Argument 's' is expected to be a non-str iterable, but is '<class 'str'>'

    """
    for arg_name, arg_value in kwargs.items():
        # Cannot import from utils -> cyclic dependency
        if not hasattr(arg_value, '__iter__') or isinstance(arg_value, (str, bytes)):
            arg_type = type(arg_value)
            raise TypeError(
                "Argument '{arg_name}' is expected to be a non-str iterable, "
                "but is '{arg_type}'".format(
                    arg_name=arg_name, arg_type=arg_type
                )
            )


def one_not_none(**kwargs: Any) -> None:
    """
    Examples:
        >>> one_not_none(arg1=None, arg2=None, arg3='passed')
        >>> one_not_none(arg1=None, arg2=None)
        Traceback (most recent call last):
        ...
        ValueError: Arguments ['arg1', 'arg2'] expects at least one passed, but all are none
    """
    if all(x is None for x in kwargs.values()):
        raise ValueError(
            "Arguments {args} expects at least one passed, but all are none".format(
                args=sorted(list(kwargs.keys()))
            )
        )


@typechecked
def one_of(possible: Iterable[Any], **kwargs: Any) -> None:
    """
    Examples:
        >>> one_of(['a', 'b', 'c'], arg='a')
        >>> one_of(['a', 'b', 'c'], arg='z')
        Traceback (most recent call last):
        ...
        ValueError: Argument 'arg' is expected to be one of ['a', 'b', 'c'], but is 'z'
        >>> one_of(['a', 'b', 'c'], arg1='a', arg2='z')
        Traceback (most recent call last):
        ...
        ValueError: Argument 'arg2' is expected to be one of ['a', 'b', 'c'], but is 'z'
    """
    for arg_name, arg_value in kwargs.items():
        if arg_value not in possible:
            raise ValueError(
                "Argument '{arg_name}' is expected to be one of {possible}, "
                "but is '{arg_value}'".format(
                    arg_name=arg_name, possible=possible, arg_value=arg_value
                )
            )


@typechecked
def subset_of(possible: Iterable[Any], **kwargs: Any) -> None:
    """
    Examples:

        >>> subset_of(['a', 'b', 'c'], arg='a')
        >>> subset_of(['a', 'b', 'c'], arg=('a', 'b'))
        >>> subset_of(['a', 'b', 'c'], arg=('a', 'b'), arg2='c')
        >>> subset_of(['a', 'b', 'c'], arg='d')
        Traceback (most recent call last):
        ...
        ValueError: Argument 'arg' is expected to be a subset of ['a', 'b', 'c'], but is 'd'
    """
    def make_list(lst: Any) -> Optional[List[Any]]:
        if lst is None:
            return None
        if isinstance(lst, list):
            return lst
        if isinstance(lst, dict):
            return [lst]
        if hasattr(lst, '__iter__') and not isinstance(lst, str):
            return list(lst)
        return [lst]

    for arg_name, arg_value in kwargs.items():
        if not set(make_list(arg_value) or set()) <= set(possible):
            raise ValueError(
                "Argument '{arg_name}' is expected to be a subset of {possible}, "
                "but is '{arg_value}'".format(
                    arg_name=arg_name, possible=possible, arg_value=arg_value
                )
            )



"""Notification related push plugins."""

from typing import Any, Dict, List, Optional

import slacker

from pnp.plugins.push import SyncPush, enveloped, parse_envelope
from pnp.typing import Envelope, Payload
from pnp.utils import make_list


class Slack(SyncPush):
    """
    Sends a message to slack. Optionally you can specify users to ping.

    See Also:
        https://pnp.readthedocs.io/en/stable/plugins/index.html#notify-slack
    """
    __REPR_FIELDS__ = ['channel', 'emoji', 'ping_users', 'username']

    DEFAULT_USERNAME = 'PnP'
    DEFAULT_EMOJI = ':robot:'

    def __init__(self, api_key: str, channel: str, username: str = DEFAULT_USERNAME,
                 emoji: str = DEFAULT_EMOJI, ping_users: Optional[List[str]] = None,
                 **kwargs: Any):
        super().__init__(**kwargs)
        self.api_key = str(api_key)
        self.channel = self._parse_channel(channel)
        self.username = self._parse_username(username)
        self.emoji = self._parse_emoji(emoji)
        self.ping_users = self._parse_ping_users(ping_users)
        self._user_cache = dict()  # type: Dict[str, str]
        self._slacker = slacker.Slacker(api_key)

    @staticmethod
    def _parse_channel(val: Any):
        channel = str(val)
        if not channel.startswith('#'):
            channel = '#' + channel
        return channel

    @staticmethod
    def _parse_ping_users(val: Any):
        return make_list(val or [])

    @staticmethod
    def _parse_username(val: Any):
        return str(val)

    @staticmethod
    def _parse_emoji(val: Any):
        return str(val)

    def _refresh_user_cache(self) -> Dict[str, int]:
        def _helper() -> Any:
            for usr in user_list:
                yield {usr['name']: usr['id']}
                yield {usr['profile']['real_name']: usr['id']}
                yield {usr['profile']['display_name']: usr['id']}

        self.logger.debug("User cache invalidation")
        user_list = self._slacker.users.list().body['members']
        self._user_cache = {k: v for d in _helper() for k, v in d.items()}

    def _lookup_users(self, ping_users: List[str]):
        if not ping_users:
            return

        refreshed = False
        if not self._user_cache:
            self._refresh_user_cache()
            refreshed = True

        for user in ping_users:
            user_id = self._user_cache.get(user, None)
            if user_id is None and not refreshed:
                # Maybe our cache is too old and some users were added in the meantime
                self._refresh_user_cache()
                refreshed = True
                user_id = self._user_cache.get(user, None)

            if user_id is None:  # Still not there...
                self.logger.warning("User not found in Slack users list: %s", user)
            else:
                yield user_id

    def _build_message_text(self, text: str, ping_users: List[str]):
        ping_user_ids = self._lookup_users(ping_users)

        for user_id in ping_user_ids:
            text = "<@{user}> {text}".format(user=user_id, text=text)

        return text

    @enveloped
    @parse_envelope('channel')
    @parse_envelope('username')
    @parse_envelope('emoji')
    @parse_envelope('ping_users')
    def _push(self, channel: str, username: str, emoji: str,  # pylint: disable=arguments-differ
              ping_users: List[str], envelope: Envelope, payload: Payload):

        text = self._build_message_text(str(payload), ping_users)
        self._slacker.chat.post_message(
            text=text,
            channel=channel,
            username=username,
            icon_emoji=emoji
        )

        return {'data': payload, **envelope} if envelope else payload
